"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev, Dario Mylonopoulos

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import moderngl
import numpy as np
from pxr import Gf, UsdGeom

from aitviewer.renderables.meshes import Meshes
from aitviewer.scene.node import Node
from aitviewer.shaders import get_chessboard_program, get_smooth_lit_with_edges_program
from aitviewer.utils import set_lights_in_program, set_material_properties
from aitviewer.utils.decorators import hooked


class Plane(Node):
    """
    Draw a plane.
    """

    def __init__(
        self,
        center,
        v1,
        v2,
        size=10.0,
        color=(0.5, 0.5, 0.5, 1.0),
        icon="\u008b",
        **kwargs,
    ):
        """
        Initializer.
        :param center: Center of the plane.
        :param v1: 3D vector lying in the plane.
        :param v2: 3D vector lying in the plane and not co-linear to `v2`.
        :param size: Size of the plane.
        :param color: Color of the plane.
        """
        super(Plane, self).__init__(color=color, icon=icon, **kwargs)
        if np.dot(v1, v2) > 0.00001:
            raise ValueError("v1 and v2 are not orthogonal.")

        self.plane_center = center
        self.v1 = v1 / np.linalg.norm(v1)
        self.v2 = v2 / np.linalg.norm(v2)
        self.size = size

        self.vertices, self.normals = self._get_renderable_data()
        self.colors = np.full((self.vertices.shape[0], 4), color)

        self.backface_culling = False

    @classmethod
    def from_normal(cls, center: np.ndarray, normal: np.ndarray, tangent: np.ndarray = None, **kwargs) -> "Plane":
        """
        Create a plane given a normal vector and optionally a tangent vector.

        :param center: Center of the plane.
        :param normal: Vector normal to the plane (doesn't have to be normalized)
        :param tangent: Optional vector tangent to the plane used to compute the plane orientation,
            must not be parallel to the normal vector, if None an axis not parallel to the normal
            will be used instead.
        :param **kwargs: arguments forwarded to the Plane constructor.
        """

        # Normalize the normal vector
        normal = normal / np.linalg.norm(normal)

        # If no tangent is given use an axis that is not perpendicular to the plane
        if tangent is None:
            if abs(np.dot(normal, np.array([1.0, 0.0, 0.0]))) < 1e-3:
                tangent = np.array([1.0, 0.0, 0.0])
            else:
                tangent = np.array([0.0, 0.0, 1.0])
        elif abs(np.dot(normal, tangent)) > 0.999:
            raise ValueError("normal and tangent are parallel")

        v1 = np.cross(normal, tangent)
        v2 = np.cross(normal, v1)
        return cls(center, v1, v2, **kwargs)

    def _get_renderable_data(self):
        p0 = self.plane_center + self.v1 * self.size - self.v2 * self.size
        p1 = self.plane_center + self.v1 * self.size + self.v2 * self.size
        p2 = self.plane_center - self.v1 * self.size + self.v2 * self.size
        p3 = self.plane_center - self.v1 * self.size - self.v2 * self.size
        normal = np.cross(self.v2, self.v1)
        normal = np.tile(normal[np.newaxis], (4, 1))
        return np.row_stack([p1, p0, p2, p3]), normal

    @Node.color.setter
    def color(self, color):
        self.material.color = color
        self.colors = np.full((self.vertices.shape[0], 4), color)
        if self.is_renderable:
            self.vbo_colors.write(self.colors.astype("f4").tobytes())

    # noinspection PyAttributeOutsideInit
    @Node.once
    def make_renderable(self, ctx):
        self.prog = get_smooth_lit_with_edges_program("lit_with_edges.glsl")
        self.vbo_vertices = ctx.buffer(self.vertices.astype("f4").tobytes())
        self.vbo_normals = ctx.buffer(self.normals.astype("f4").tobytes())
        self.vbo_colors = ctx.buffer(self.colors.astype("f4").tobytes())

        self.vao = ctx.vertex_array(
            self.prog,
            [
                (self.vbo_vertices, "3f4 /v", "in_position"),
                (self.vbo_normals, "3f4 /v", "in_normal"),
                (self.vbo_colors, "4f4 /v", "in_color"),
            ],
        )

    def render(self, camera, **kwargs):
        self.prog["norm_coloring"].value = False
        self.prog["win_size"].value = kwargs["window_size"]

        self.set_camera_matrices(self.prog, camera, **kwargs)
        set_lights_in_program(
            self.prog,
            kwargs["lights"],
            kwargs["shadows_enabled"],
            kwargs["ambient_strength"],
        )
        set_material_properties(self.prog, self.material)
        self.receive_shadow(self.prog, **kwargs)
        self.vao.render(moderngl.TRIANGLE_STRIP)

    @hooked
    def release(self):
        if self.is_renderable:
            self.vbo_vertices.release()
            self.vbo_normals.release()
            self.vbo_colors.release()
            self.vao.release()


class ChessboardPlane(Node):
    """A plane that is textured like a chessboard."""

    def __init__(
        self,
        side_length,
        n_tiles,
        color1=(0.0, 0.0, 0.0, 1.0),
        color2=(1.0, 1.0, 1.0, 1.0),
        plane="xz",
        height=0.0,
        tiling=True,
        icon="\u008b",
        **kwargs,
    ):
        """
        Initializer.
        :param side_length: Length of one side of the plane.
        :param n_tiles: Number of tiles for the chessboard pattern.
        :param color1: First color of the chessboard pattern.
        :param color2: Second color of the chessboard pattern.
        :param plane: In which plane the chessboard lies. Allowed are 'xz', 'xy', 'yz'.
        :param height: The height of the plane.
        :param kwargs: Remaining kwargs.
        """
        assert plane in ["xz", "xy", "yz"]
        super(ChessboardPlane, self).__init__(icon=icon, **kwargs)
        self.side_length = side_length
        self.n_tiles = n_tiles
        self.c1 = np.array(color1)
        self.c2 = np.array(color2)
        self.plane = plane
        self.height = height
        self.tiling = tiling

        if plane == "xz":
            v1 = np.array([1, 0, 0], dtype=np.float32)
            v2 = np.array([0, 0, 1], dtype=np.float32)
        elif plane == "xy":
            v1 = np.array([1, 0, 0], dtype=np.float32)
            v2 = np.array([0, 1, 0], dtype=np.float32)
        else:
            # plane == "yz"
            v1 = np.array([0, 1, 0], dtype=np.float32)
            v2 = np.array([0, 0, 1], dtype=np.float32)

        self.vertices, self.normals, self.uvs = self._get_renderable_data(v1, v2, side_length)
        self.backface_culling = False

    def _get_renderable_data(self, v1, v2, size):
        p0 = v1 * (size * 0.5) - v2 * (size * 0.5)
        p1 = v1 * (size * 0.5) + v2 * (size * 0.5)
        p2 = -v1 * (size * 0.5) + v2 * (size * 0.5)
        p3 = -v1 * (size * 0.5) - v2 * (size * 0.5)

        normals = np.tile(np.cross(v2, v1), (4, 1))

        uvs = np.array([0, 0, 0, 1, 1, 0, 1, 1], dtype=np.float32)

        return np.row_stack([p1, p0, p2, p3]), normals, uvs

    # noinspection PyAttributeOutsideInit
    @Node.once
    def make_renderable(self, ctx):
        self.prog = get_chessboard_program()
        self.vbo_vertices = ctx.buffer(self.vertices.astype("f4").tobytes())
        self.vbo_normals = ctx.buffer(self.normals.astype("f4").tobytes())
        self.vbo_uvs = ctx.buffer(self.uvs.astype("f4").tobytes())

        self.vao = ctx.vertex_array(
            self.prog,
            [
                (self.vbo_vertices, "3f4 /v", "in_position"),
                (self.vbo_normals, "3f4 /v", "in_normal"),
                (self.vbo_uvs, "2f4 /v", "in_uv"),
            ],
        )

    def render(self, camera, **kwargs):
        self.prog["color_1"].value = (self.c1[0], self.c1[1], self.c1[2], self.c1[3])
        self.prog["color_2"].value = (self.c2[0], self.c2[1], self.c2[2], self.c2[3])
        self.prog["n_tiles"].value = self.n_tiles
        self.prog["tiling_enabled"].value = self.tiling

        self.set_camera_matrices(self.prog, camera, **kwargs)
        self.receive_shadow(self.prog, **kwargs)

        set_lights_in_program(
            self.prog,
            kwargs["lights"],
            kwargs["shadows_enabled"],
            kwargs["ambient_strength"],
        )
        set_material_properties(self.prog, self.material)

        self.vao.render(moderngl.TRIANGLE_STRIP)

    @property
    def bounds(self):
        return self.get_bounds(self.vertices)

    @property
    def current_bounds(self):
        return self.bounds

    def gui(self, imgui):
        _, self.c1 = imgui.color_edit4("Color 1##color{}'".format(self.unique_name), *self.c1)
        _, self.c2 = imgui.color_edit4("Color 2##color{}'".format(self.unique_name), *self.c2)
        _, self.tiling = imgui.checkbox("Toggle Tiling", self.tiling)
        _, self.n_tiles = imgui.drag_int("Number of tiles", self.n_tiles, 1.0, 1, 200)

    def export_usd(self, stage, usd_path: str, directory: str = None, verbose=False):
        usd_path = f"{usd_path}/{self.name}_{self.uid:03}".replace(" ", "_")

        # Transform.
        xform = UsdGeom.Xform.Define(stage, usd_path)
        a_xform = xform.AddTransformOp()
        a_xform.Set(Gf.Matrix4d(self.get_local_transform().astype(np.float64).T))

        # Geometry.
        mesh = UsdGeom.Mesh.Define(stage, usd_path + "/" + self.name.replace(" ", "_"))
        mesh.CreatePointsAttr(self.vertices)
        mesh.CreateFaceVertexCountsAttr(np.array([4]))
        mesh.CreateFaceVertexIndicesAttr([0, 1, 3, 2])
        self._export_usd_recursively(stage, usd_path, directory, verbose)


class Chessboard(Node):
    """A plane that is textured like a chessboard."""

    def __init__(
        self,
        side_length,
        n_tiles,
        color1=(0.0, 0.0, 0.0, 1.0),
        color2=(1.0, 1.0, 1.0, 1.0),
        plane="xz",
        height=0.0,
        tiling=True,
        **kwargs,
    ):
        """
        Initializer.
        :param side_length: Length of one side of the plane.
        :param n_tiles: Number of tiles for the chessboard pattern.
        :param color1: First color of the chessboard pattern.
        :param color2: Second color of the chessboard pattern.
        :param plane: In which plane the chessboard lies. Allowed are 'xz', 'xy', 'yz'.
        :param height: The height of the plane.
        :param kwargs: Remaining kwargs.
        """
        assert plane in ["xz", "xy", "yz"]
        super(Chessboard, self).__init__(**kwargs)
        self.side_length = side_length
        self.n_tiles = n_tiles
        self.c1 = np.array(color1)
        self.c2 = np.array(color2)
        self.plane = plane
        self.height = height
        self.tiling = tiling

        vs, fs, fc, c1_idxs, c2_idxs = self._construct_board()
        self.fcs_tiled = fc
        self.c1_idxs = c1_idxs
        self.c2_idxs = c2_idxs

        self.mesh = Meshes(vs, fs, face_colors=fc)
        self.mesh.backface_culling = False
        self.add(self.mesh, show_in_hierarchy=False)

    # noinspection PyAttributeOutsideInit
    def _construct_board(self):
        """Construct the chessboard mesh."""
        vertices = []
        faces = []
        face_colors = []
        c1_idxs = []  # Store indices into face_colors containing color 1.
        c2_idxs = []  # Store indices into face_colors containing color 2.
        tl = self.side_length / self.n_tiles
        dim1 = "xyz".index(self.plane[0])
        dim2 = "xyz".index(self.plane[1])
        up = "xyz".index("xyz".replace(self.plane[0], "").replace(self.plane[1], ""))

        for r in range(self.n_tiles):
            for c in range(self.n_tiles):
                v0 = np.zeros([3])
                v0[dim1] = r * tl
                v0[dim2] = c * tl

                v1 = np.zeros([3])
                v1[dim1] = (r + 1) * tl
                v1[dim2] = c * tl

                v2 = np.zeros([3])
                v2[dim1] = (r + 1) * tl
                v2[dim2] = (c + 1) * tl

                v3 = np.zeros([3])
                v3[dim1] = r * tl
                v3[dim2] = (c + 1) * tl

                vertices.extend([v0, v1, v2, v3])

                # Need counter-clock-wise ordering
                faces.append([len(vertices) - 4, len(vertices) - 1, len(vertices) - 3])
                faces.append([len(vertices) - 3, len(vertices) - 1, len(vertices) - 2])

                if r % 2 == 0 and c % 2 == 0:
                    c = self.c1
                    fc_idxs = c1_idxs
                elif r % 2 == 0 and c % 2 != 0:
                    c = self.c2
                    fc_idxs = c2_idxs
                elif r % 2 != 0 and c % 2 != 0:
                    c = self.c1
                    fc_idxs = c1_idxs
                else:
                    c = self.c2
                    fc_idxs = c2_idxs
                face_colors.append(c)
                face_colors.append(c)
                fc_idxs.extend([len(face_colors) - 2, len(face_colors) - 1])

        vs = np.stack(vertices)
        vs = vs - np.mean(vertices, axis=0, keepdims=True)
        vs[:, up] = self.height
        fs = np.stack(faces)
        cs = np.stack(face_colors)

        return vs, fs, cs, c1_idxs, c2_idxs

    def _update_colors(self):
        self.fcs_tiled[self.c1_idxs] = self.c1
        self.fcs_tiled[self.c2_idxs] = self.c2 if self.tiling else self.c1
        self.mesh.face_colors = self.fcs_tiled

    def gui(self, imgui):
        u, c1 = imgui.color_edit4("Color 1##color{}'".format(self.unique_name), *self.c1)
        if u:
            self.c1 = c1
            self._update_colors()

        u, c2 = imgui.color_edit4("Color 2##color{}'".format(self.unique_name), *self.c2)
        if u:
            self.c2 = c2
            self._update_colors()

        u, self.tiling = imgui.checkbox("Toggle Tiling", self.tiling)
        if u:
            self._update_colors()



from ..message import Message
from ..node import RemoteNode


class RemoteMeshes(RemoteNode):
    MESSAGE_TYPE = Message.MESHES

    def __init__(self, viewer, vertices, faces, **kwargs):
        """
        This initializer takes a RemoteViewer object and all other arguments are forwarded
        to the Meshes constructor on the remote Viewer.
        See the Meshes class for more information about parameters.

        :param viewer: a RemoteViewer object that will be used to send this node.
        :param vertices: Set of 3D coordinates as a np array of shape (N, V, 3) or (V, 3).
        """
        super().__init__(
            viewer,
            vertices=vertices,
            faces=faces,
            **kwargs,
        )

    def add_frames(self, vertices):
        """
        Add frames to the remote Meshes node by adding new vertices.

        :param vertices: Set of 3D coordinates as a np array of shape (N, V, 3) or (V, 3).
        """
        return super().add_frames(vertices=vertices)

    def update_frames(self, vertices, frames):
        """
        Update frames of the remote Meshes node by adding new vertices.

        :param vertices: Set of 3D coordinates as a np array of shape (N, V, 3) or (V, 3).
        :param frames: a list of integer frame indices of size N or a single integer frame index.
        """
        return super().update_frames(vertices=vertices, frames=frames)



import asyncio
import pickle
import queue
import threading
from typing import Dict, Tuple

import websockets

from aitviewer.models.smpl import SMPLLayer
from aitviewer.remote.message import Message
from aitviewer.renderables.arrows import Arrows
from aitviewer.renderables.lines import Lines
from aitviewer.renderables.meshes import Meshes
from aitviewer.renderables.rigid_bodies import RigidBodies
from aitviewer.renderables.smpl import SMPLSequence
from aitviewer.renderables.spheres import Spheres
from aitviewer.scene.node import Node


class ViewerServer:
    def __init__(self, viewer, port):
        """
        Initializer.
        :param viewer: the viewer for which the server processes messages.
        :param port: the TCP port number to which the websocket for listening to incoming
          connections will be bound.
        """
        self.viewer = viewer

        # Dictionary mapping from the remote node uid to the local node uid
        self.remote_to_local_id = {}

        # A queue of messages received by the server. This is used to transfer
        # received messages from the server thread to the main thread.
        self.queue = queue.Queue()

        # A map of connections that are currently open. Each entry maps
        # the remote address (host, port) to a websocket.
        self.connections: Dict[Tuple[str, str], websockets.server.WebSocketServerProtocol] = {}

        # Entry point of server thread
        def entry():
            # Called whenever a new connection is enstablished.
            async def serve(websocket):
                addr = websocket.remote_address
                self.connections[addr] = websocket

                print(f"New connection: {addr[0]}:{addr[1]}")
                try:
                    # Message loop.
                    # This loop is exited whenever the connection is dropped
                    # which causes and exception to be raised.
                    async for message in websocket:
                        data = pickle.loads(message)
                        # Equeue data for the main thread to process.
                        self.queue.put_nowait((addr, data))
                    await websocket.close()
                except Exception as e:
                    print(f"Except {e}")
                del self.connections[addr]
                print(f"Connection closed: {addr[0]}:{addr[1]}")

            # Async entry point of the main thread.
            async def main():
                async with websockets.serve(serve, "0.0.0.0", port, max_size=None):
                    await self.stop

            asyncio.set_event_loop(self.loop)
            self.loop.run_until_complete(main())

        self.loop = asyncio.new_event_loop()
        self.stop = self.loop.create_future()
        # daemon = true means that the thread is abruptly stopped once the main thread exits.
        self.thread = threading.Thread(target=entry, daemon=True)
        self.thread.start()

    def process_messages(self):
        """Processes all messages received since the last time this method was called."""
        while not self.queue.empty():
            client, msg = self.queue.get_nowait()
            # Call process_message on the viewer so that subclasses of a viewer can intercept messages.
            # By default this will end up calling self.process_message()
            self.viewer.process_message(msg["type"], msg["uid"], msg["args"], msg["kwargs"], client)

    def process_message(self, type: Message, remote_uid: int, args: list, kwargs: dict, client: Tuple[str, str]):
        """
        Default processing of messages.

        :param type: the type of the message.
        :param remote_uid: the remote id of the node that this message refers to.
        :param args: positional arguments received with the message.
        :param kwargs: keyword arguments received with the message.
        :param client: a tuple (ip, port) describing the address of the client
            that sent this message.
        """

        def add(client, remote_uid, args, kwargs, type):
            n = type(*args, **kwargs)
            self.viewer.scene.add(n)
            self.remote_to_local_id[(client, remote_uid)] = n.uid

        if type == Message.NODE:
            add(client, remote_uid, args, kwargs, Node)

        elif type == Message.MESHES:
            add(client, remote_uid, args, kwargs, Meshes)

        elif type == Message.SPHERES:
            add(client, remote_uid, args, kwargs, Spheres)

        elif type == Message.LINES:
            add(client, remote_uid, args, kwargs, Lines)

        elif type == Message.ARROWS:
            add(client, remote_uid, args, kwargs, Arrows)

        elif type == Message.RIGID_BODIES:
            add(client, remote_uid, args, kwargs, RigidBodies)

        elif type == Message.SMPL:
            layer_arg_names = {"model_type", "gender", "num_betas"}
            layer_kwargs = {k: v for k, v in kwargs.items() if k in layer_arg_names}
            layer = SMPLLayer(**layer_kwargs)

            sequence_kwargs = {k: v for k, v in kwargs.items() if k not in layer_arg_names}
            n = SMPLSequence(*args, smpl_layer=layer, **sequence_kwargs)
            self.viewer.scene.add(n)
            self.remote_to_local_id[(client, remote_uid)] = n.uid

        elif type == Message.DELETE:
            node: Node = self.get_node_by_remote_uid(remote_uid, client)
            if node and node.parent:
                node.parent.remove(node)

        elif type == Message.UPDATE_FRAMES:
            node: Node = self.get_node_by_remote_uid(remote_uid, client)
            if node:
                node.update_frames(*args, **kwargs)

        elif type == Message.ADD_FRAMES:
            node: Node = self.get_node_by_remote_uid(remote_uid, client)
            if node:
                node.add_frames(*args, **kwargs)

        elif type == Message.REMOVE_FRAMES:
            node: Node = self.get_node_by_remote_uid(remote_uid, client)
            if node:
                node.remove_frames(*args, **kwargs)

        elif type == Message.SET_FRAME:
            if not self.viewer.run_animations:
                self.viewer.scene.current_frame_id = args[0]

        elif type == Message.NEXT_FRAME:
            if not self.viewer.run_animations:
                self.viewer.scene.next_frame()

        elif type == Message.PREVIOUS_FRAME:
            if not self.viewer.run_animations:
                self.viewer.scene.previous_frame()

    async def _async_send(self, msg, client):
        data = pickle.dumps(msg)
        if client is None:
            websockets.broadcast(self.connections.values(), data)
        else:
            websocket = self.connections.get(client)
            if websocket is not None:
                await websocket.send(msg)

    def send_message(self, msg, client: Tuple[str, str] = None):
        """
        Send a message to a single client or to all connected clients.

        :param msg: a python object that is serialized with pickle and sent to the client.
        :param client: a tuple (host, port) representing the client to which to send the message,
            if None the message is sent to all connected clients.
        """
        # Send a message by adding a send coroutine to the thread's loop and wait for it to complete.
        asyncio.run_coroutine_threadsafe(self._async_send(msg, client), self.loop).result()

    def get_node_by_remote_uid(self, remote_uid: int, client: Tuple[str, str]):
        """
        Returns the Node corresponding to the remote uid and client passed in.

        :param remote_uid: the remote uid to look up.
        :param client: the client that created the node, this is the value of the 'client'
            parameter that was passed to process_message() when the message was received.
        :return: Node corresponding to the remote uid.
        """
        return self.viewer.scene.get_node_by_uid(self.remote_to_local_id.get((client, remote_uid), None))

    def close(self):
        self.loop.call_soon_threadsafe(self.stop.set_result, None)
        self.thread.join()


# If this module is invoke directly it starts an empty viewer
# with the server functionality enabled.
if __name__ == "__main__":
    from aitviewer.configuration import CONFIG as C
    from aitviewer.viewer import Viewer

    C.update_conf({"server_enabled": True})
    v = Viewer()
    v.scene.floor.enabled = False
    v.run()



import enum


class Message(enum.Enum):
    """Enumeration for the type of message."""

    # Messages used to create nodes on the remote viewer.
    NODE = 1
    MESHES = 2
    SPHERES = 3
    LINES = 4
    ARROWS = 5
    RIGID_BODIES = 6
    SMPL = 10

    # Messages used to modify existing nodes on the remote viewer.
    DELETE = 100
    ADD_FRAMES = 101
    UPDATE_FRAMES = 102
    REMOVE_FRAMES = 103

    # Built-in uitilities to interact with the viewer.
    SET_FRAME = 200
    NEXT_FRAME = 201
    PREVIOUS_FRAME = 202

    # All values greater than or equal to Message.USER_MESSAGE are not used internally by the viewer
    # and can be safely used to send custom messages.
    USER_MESSAGE = 10000


def make_message(type, uid, args, kwargs):
    msg = {
        "type": type,
        "uid": uid,
        "args": args,
        "kwargs": kwargs,
    }
    return msg



from ..message import Message
from ..node import RemoteNode


class RemoteLines(RemoteNode):
    MESSAGE_TYPE = Message.LINES

    def __init__(self, viewer, lines, **kwargs):
        """
        This initializer takes a RemoteViewer object and all other arguments are forwarded
        to the Lines constructor on the remote Viewer.
        See the Lines class for more information about parameters.

        :param viewer: a RemoteViewer object that will be used to send this node.
        :param lines: Set of 3D coordinates as a np array of shape (N, L, 3) or (L, 3).
        """
        super().__init__(
            viewer,
            lines=lines,
            **kwargs,
        )

    def add_frames(self, lines):
        """
        Add frames to the remote Lines node by adding new lines.

        :param lines: Set of 3D coordinates as a np array of shape (N, L, 3) or (L, 3).
        """
        return super().add_frames(lines=lines)

    def update_frames(self, lines, frames):
        """
        Update frames of the remote Lines node by updating the origins and tips.

        :param lines: Set of 3D coordinates as a np array of shape (N, L, 3) or (L, 3).
        :param frames: a list of integer frame indices of size N or a single integer frame index.
        """
        return super().update_frames(lines=lines, frames=frames)



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev, Dario Mylonopoulos

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import collections
import os

import torch
import torch.nn.functional as F

try:
    from star.config import cfg
    from star.pytorch.star import STAR
except Exception as e:
    raise ImportError(f"Cannot import STAR. Please run `pip install git+https://github.com/ahmedosman/STAR.git`\n{e}")

from aitviewer.configuration import CONFIG as C
from aitviewer.utils.so3 import aa2rot_torch as aa2rot
from aitviewer.utils.so3 import rot2aa_torch as rot2aa


class STARLayer(STAR):
    """Wraps the publicly available STAR model to match SMPLX model interface"""

    def __init__(self, gender="male", num_betas=300, device=None, dtype=None):
        """
        Initializer.
        :param gender: Which gender to load.
        :param num_betas: Number of shape components. Keep in mind STAR and SMPL shape components are not the same
        (i.e. the first 10 components of STAR cannot simply be assigned from the first 10 components of SMPL)
        :param device: CPU or GPU.
        :param dtype: The pytorch floating point data type.
        """
        # Configure STAR model before initializing
        cfg.data_type = dtype if dtype is not None else C.f_precision
        cfg.path_male_star = os.path.join(C.star_models, "male/model.npz")
        cfg.path_female_star = os.path.join(C.star_models, "female/model.npz")
        cfg.path_neutral_star = os.path.join(C.star_models, "neutral/model.npz")

        super(STARLayer, self).__init__(gender=gender, num_betas=num_betas)

        self.device = device if device is not None else C.device
        self.model_type = "star"
        self._parents = None
        self._children = None

    @property
    def parents(self):
        """Return how the joints are connected in the kinematic chain where parents[i, 0] is the parent of
        joint parents[i, 1]."""
        if self._parents is None:
            self._parents = self.kintree_table.transpose(0, 1).cpu().numpy()
        return self._parents

    @property
    def joint_children(self):
        """Return the children of each joint in the kinematic chain."""
        if self._children is None:
            self._children = collections.defaultdict(list)
            for bone in self.parents:
                if bone[0] != -1:
                    self._children[bone[0]].append(bone[1])
        return self._children

    def skeletons(self):
        """Return how the joints are connected in the kinematic chain where skeleton[0, i] is the parent of
        joint skeleton[1, i]."""
        kintree_table = self.kintree_table
        kintree_table[:, 0] = -1
        return {
            "all": kintree_table,
            "body": kintree_table[:, : self.n_joints_body + 1],
        }

    @property
    def n_joints_body(self):
        return self.parent.shape[0]

    @property
    def n_joints_total(self):
        return self.n_joints_body + 1

    def forward(self, poses_body, betas=None, poses_root=None, trans=None, normalize_root=False):
        """
        forwards the model
        :param poses_body: Pose parameters.
        :param poses_root: Pose parameters for the root joint.
        :param beta: Beta parameters.
        :param trans: Root translation.
        :param normalize_root: Makes poses relative to the root joint (useful for globally rotated captures).
        :return: Deformed surface vertices, transformed joints
        """
        poses, betas, trans = self.preprocess(poses_body, betas, poses_root, trans, normalize_root)

        # STAR repo currently hardcodes floats.
        v = super().forward(pose=poses.float(), betas=betas.float(), trans=trans.float())
        J = v.J_transformed
        return v, J

    def preprocess(self, poses_body, betas=None, poses_root=None, trans=None, normalize_root=False):
        batch_size = poses_body.shape[0]

        if poses_root is None:
            poses_root = torch.zeros([batch_size, 3]).to(dtype=poses_body.dtype, device=self.device)
        if trans is None:
            # If we don't supply the root translation explicitly, it falls back to using self.bm.trans
            # which might not be zero since it is a trainable param that can get updated.
            trans = torch.zeros([batch_size, 3]).to(dtype=poses_body.dtype, device=self.device)

        if normalize_root:
            # Make everything relative to the first root orientation.
            root_ori = aa2rot(poses_root)
            first_root_ori = torch.inverse(root_ori[0:1])
            root_ori = torch.matmul(first_root_ori, root_ori)
            poses_root = rot2aa(root_ori)
            trans = torch.matmul(first_root_ori.unsqueeze(0), trans.unsqueeze(-1)).squeeze()
            trans = trans - trans[0:1]

        poses = torch.cat((poses_root, poses_body), dim=1)

        if betas is None:
            betas = torch.zeros([batch_size, self.num_betas]).to(dtype=poses_body.dtype, device=self.device)

        # Batch shapes if they don't match batch dimension.
        if betas.shape[0] != batch_size:
            betas = betas.repeat(batch_size, 1)

        # Lower bound betas
        if betas.shape[1] < self.num_betas:
            betas = torch.nn.functional.pad(betas, [0, self.num_betas - betas.shape[1]])

        # Upper bound betas
        betas = betas[:, : self.num_betas]

        return poses, betas, trans



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev, Dario Mylonopoulos

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
import subprocess

import numpy as np
import torch
from scipy.interpolate import CubicSpline

from aitviewer.utils.so3 import aa2rot_torch as aa2rot
from aitviewer.utils.so3 import rot2aa_torch as rot2aa


def to_torch(x, dtype, device):
    if x is None:
        return None
    if isinstance(x, np.ndarray):
        return torch.from_numpy(x).to(dtype=dtype, device=device)
    return x.to(dtype=dtype, device=device)


def to_numpy(x):
    if isinstance(x, np.ndarray):
        return x
    return x.detach().cpu().numpy()


def get_video_paths(video_path, ensure_no_overwrite=True):
    is_webm = video_path.endswith(".webm")
    is_mp4 = video_path.endswith(".mp4")
    is_gif = video_path.endswith(".gif")
    if not (is_mp4 or is_gif or is_webm):
        video_path += ".mp4"
        is_mp4 = True

    suffix = os.path.splitext(video_path)[1]

    dir_of_file = os.path.dirname(os.path.abspath(video_path))
    if not os.path.exists(dir_of_file):
        os.makedirs(dir_of_file)

    # Make sure we don't override an existing video.
    if ensure_no_overwrite:
        counter = 0
        video_path_candidate = video_path.replace(suffix, f"_{counter}{suffix}")
        while os.path.exists(video_path_candidate):
            counter += 1
            video_path_candidate = video_path.replace(suffix, f"_{counter}{suffix}")
    else:
        video_path_candidate = video_path

    if is_mp4 or is_webm:
        return video_path_candidate, None, is_gif
    else:
        video_path_temp = video_path_candidate.replace(suffix, f"_temp.mp4")
        return video_path_temp, video_path_candidate, is_gif


def video_to_gif(path_mp4, path_gif, remove=False):
    command = [
        "ffmpeg",
        "-i",
        path_mp4,
        "-y",
        "-vf",
        "split[s0][s1];[s0]palettegen[p];[s1][p]paletteuse",
        path_gif,
    ]

    with open(os.devnull, "w") as FNULL:
        subprocess.Popen(command, stdout=FNULL, stderr=FNULL).wait()

    if remove:
        os.remove(path_mp4)


def images_to_video(
    frame_dir,
    video_path,
    frame_format="frame_%06d.png",
    input_fps=60,
    output_fps=60,
    start_frame=0,
):
    """Convert the rendered images into a video. The video path format determines whether this will be rendered as
    a GIF or an MP4 (default)."""
    if not os.path.exists(frame_dir):
        raise ValueError(f"Could not find directory containing frames {frame_dir}")

    path_mp4, path_gif, is_gif = get_video_paths(video_path)

    print("Rendering to video {}".format(os.path.abspath(path_mp4)))

    # Create mp4 from frames.
    command = [
        "ffmpeg",
        "-framerate",
        str(input_fps),  # must be this early in the command, otherwise it is not applied.
        "-start_number",
        str(start_frame),
        "-i",
        os.path.join(frame_dir, frame_format),
        "-c:v",
        "libx264",
        "-preset",
        "slow",
        "-profile:v",
        "high",
        "-level:v",
        "4.0",
        "-pix_fmt",
        "yuv420p",
        "-vf",
        "pad=ceil(iw/2)*2:ceil(ih/2)*2",  # Avoid error when image res is not divisible by 2.
        "-r",
        str(output_fps),
        "-y",
        path_mp4,
    ]

    with open(os.devnull, "w") as FNULL:
        subprocess.Popen(command, stdout=FNULL, stderr=FNULL).wait()

    # If GIF we first create a temporary mp4 and then convert that to GIF to reduce palette artifacts.
    if is_gif:
        video_to_gif(path_mp4, path_gif)


def interpolate_positions(positions, ts_in, ts_out):
    """
    Interpolate positions given at timestamps `ts_in` to timestamps given at `ts_out` with a cubic spline.
    :param positions: A numpy array of shape (F, N, 3), i.e. in angle-axis form.
    :param ts_in: Timestamps corresponding to the given positions, len(ts_in) == F
    :param ts_out: The desired output timestamps.
    :return: A numpy array of shape (len(ts_out), N, 3).
    """
    cs = CubicSpline(ts_in, positions, axis=0)
    new_positions = cs(ts_out)
    return new_positions


def resample_positions(positions, fps_in, fps_out):
    """
    Resample 3D positions from `fps_in` to `fps_out`.
    :param positions: A numpy array of shape (F, ...).
    :param fps_in: The frequency of the input sequence.
    :param fps_out: The desired output frequency.
    :return: A numpy array of shape (F', ...) where F is adjusted according to the new fps.
    """
    n_frames = positions.shape[0]
    assert n_frames > 1, "Resampling with one data point does not make sense."
    duration = n_frames / fps_in
    ts_in = np.arange(0, duration, 1 / fps_in)[:n_frames]
    ts_out = np.arange(0, duration, 1 / fps_out)
    return interpolate_positions(positions, ts_in, ts_out)


def compute_vertex_and_face_normals_torch(vertices, faces, vertex_faces, normalize=False):
    """
    Compute (unnormalized) vertex normals for the given vertices.
    :param vertices: A tensor of shape (N, V, 3).
    :param faces: A tensor of shape (F, 3) indexing into `vertices`.
    :param vertex_faces: A tensor of shape (V, MAX_VERTEX_DEGREE) that lists the face IDs each vertex is a part of.
    :param normalize: Whether to make the normals unit length or not.
    :return: The vertex and face normals as tensors of shape (N, V, 3) and (N, F, 3) respectively.
    """
    vs = vertices[:, faces.to(dtype=torch.long)]
    face_normals = torch.cross(vs[:, :, 1] - vs[:, :, 0], vs[:, :, 2] - vs[:, :, 0], dim=-1)  # (N, F, 3)

    ns_all_faces = face_normals[:, vertex_faces]  # (N, V, MAX_VERTEX_DEGREE, 3)
    ns_all_faces[:, vertex_faces == -1] = 0.0
    vertex_degrees = (vertex_faces > -1).sum(dim=-1).to(dtype=ns_all_faces.dtype)
    vertex_normals = ns_all_faces.sum(dim=-2) / vertex_degrees[None, :, None]  # (N, V, 3)

    if normalize:
        face_normals = face_normals / torch.norm(face_normals, dim=-1).unsqueeze(-1)
        vertex_normals = vertex_normals / torch.norm(vertex_normals, dim=-1).unsqueeze(-1)

    return vertex_normals, face_normals


def compute_vertex_and_face_normals(vertices, faces, vertex_faces, normalize=False):
    """
    Compute (unnormalized) vertex normals for the given vertices. This is a rather expensive operation despite it being
    fully optimized with numpy. For a typical SMPL mesh this can take 5-10ms per call. Not normalizing the resulting
    normals speeds things up considerably, so we are omitting this as OpenGL shaders can do this much more efficiently.
    For more speedup, consider caching the result of this function.

    :param vertices: A numpy array of shape (N, V, 3).
    :param faces: A numpy array of shape (F, 3) indexing into `vertices`.
    :param vertex_faces: A numpy array of shape (V, MAX_VERTEX_DEGREE) that lists the face IDs each vertex is a part of.
    :param normalize: Whether to normalize the normals or not.
    :return: The vertex and face normals as a np arrays of shape (N, V, 3) and (N, F, 3) respectively.
    """
    vs = vertices[:, faces]
    face_normals = np.cross(vs[:, :, 1] - vs[:, :, 0], vs[:, :, 2] - vs[:, :, 0], axis=-1)  # (N, F, 3)

    ns_all_faces = face_normals[:, vertex_faces]  # (N, V, MAX_VERTEX_DEGREE, 3)
    ns_all_faces[:, vertex_faces == -1] = 0.0
    vertex_degrees = np.sum(vertex_faces > -1, axis=-1)
    vertex_normals = np.sum(ns_all_faces, axis=-2) / vertex_degrees[np.newaxis, :, np.newaxis]  # (N, V, 3)

    if normalize:
        face_normals = face_normals / np.linalg.norm(face_normals, axis=-1)[..., np.newaxis]
        vertex_normals = vertex_normals / np.linalg.norm(vertex_normals, axis=-1)[..., np.newaxis]

    return vertex_normals, face_normals


def compute_vertex_and_face_normals_sparse(vertices, faces, vertex_faces_sparse, normalize=False):
    """
    Compute (unnormalized) vertex normals for the given vertices. This is a rather expensive operation despite it being
    fully optimized with numpy. For a typical SMPL mesh this can take 5-10ms per call. Not normalizing the resulting
    normals speeds things up considerably, so we are omitting this as OpenGL shaders can do this much more efficiently.
    For more speedup, consider caching the result of this function.

    :param vertices: A numpy array of shape (N, V, 3).
    :param faces: A numpy array of shape (F, 3) indexing into `vertices`.
    :param vertex_faces: A scipy sparse adjacency matrix shape (V, F), each row contains a one for all faces
        that are adjacent to the corresponding vertex and a 0 otherwise.
    :param normalize: Whether to normalize the normals or not.
    :return: The vertex and face normals as a np arrays of shape (N, V, 3) and (N, F, 3) respectively.
    """
    triangles = vertices[:, faces]
    vectors = np.diff(triangles, axis=2)
    crosses = np.cross(vectors[:, :, 0], vectors[:, :, 1])

    normals = np.moveaxis(crosses, 0, 2)
    normals = normals.reshape((normals.shape[0], -1))
    vn = vertex_faces_sparse.dot(normals)
    vn = vn.reshape((vn.shape[0], 3, -1))
    vn = np.moveaxis(vn, 2, 0)

    if normalize:
        crosses = crosses / np.linalg.norm(crosses, axis=-1)[..., np.newaxis]
        vn = vn / np.linalg.norm(vn, axis=-1)[..., np.newaxis]

    return vn, crosses


def spherical_coordinates_from_direction(v, degrees=False):
    """
    Converts from a normalized direction vector to spherical coordinates.

    :param v: direction vector as np array of shape (3)
    :param degrees: if True returned spherical coordinates are expressed in degrees.
    :return: elevation and azimuthal angles
    """
    theta = np.arcsin(v[1])
    phi = np.arctan2(v[2], v[0])
    phi = np.remainder(phi, np.pi * 2.0)
    if degrees:
        return np.degrees(theta), np.degrees(phi)
    else:
        return theta, phi


def direction_from_spherical_coordinates(theta, phi, degrees=False):
    """
    Converts from spherical coordinates to a direction vector.

    :param theta: elevation angle from -PI/2 to PI/2, 0 represents the horizon and PI/2 and -PI/2 the +y and -y directions respectively
    :param phi: azimuthal angle from 0 to 2 * PI, 0 and PI represents the +x and -x directions respectively
        PI/2 and 3/2P the +z and -z respectively.
    :param degrees: if True theta and phi must be expressed in degrees instead.
    :return: The normalized direction as a np array of shape (3)
    """
    if degrees:
        theta, phi = np.radians(theta), np.radians(phi)
    sin_theta = np.sin(theta)
    cos_theta = np.cos(theta)
    sin_phi = np.sin(phi)
    cos_phi = np.cos(phi)

    x = cos_theta * cos_phi
    y = sin_theta
    z = cos_theta * sin_phi

    return np.array([x, y, z])


def set_lights_in_program(prog, lights, shadows_enabled, ambient_strength):
    """Set program lighting from scene lights"""
    for i, light in enumerate(lights):
        prog[f"dirLights[{i}].direction"].value = tuple(light.direction)
        prog[f"dirLights[{i}].color"].value = light.light_color
        prog[f"dirLights[{i}].strength"].value = light.strength if light.enabled else 0.0
        prog[f"dirLights[{i}].shadow_enabled"].value = shadows_enabled and light.shadow_enabled
    prog["ambient_strength"] = ambient_strength


def set_material_properties(prog, material):
    prog["diffuse_coeff"].value = material.diffuse
    prog["ambient_coeff"].value = material.ambient


def compute_union_of_bounds(nodes):
    if len(nodes) == 0:
        return np.zeros((3, 2))

    bounds = np.array([[np.inf, np.NINF], [np.inf, np.NINF], [np.inf, np.NINF]])
    for n in nodes:
        child = n.bounds
        bounds[:, 0] = np.minimum(bounds[:, 0], child[:, 0])
        bounds[:, 1] = np.maximum(bounds[:, 1], child[:, 1])
    return bounds


def compute_union_of_current_bounds(nodes):
    if len(nodes) == 0:
        return np.zeros((3, 2))

    bounds = np.array([[np.inf, np.NINF], [np.inf, np.NINF], [np.inf, np.NINF]])
    for n in nodes:
        child = n.current_bounds
        bounds[:, 0] = np.minimum(bounds[:, 0], child[:, 0])
        bounds[:, 1] = np.maximum(bounds[:, 1], child[:, 1])
    return bounds


def local_to_global(poses, parents, output_format="aa", input_format="aa"):
    """
    Convert relative joint angles to global ones by unrolling the kinematic chain.
    :param poses: A tensor of shape (N, N_JOINTS*3) defining the relative poses in angle-axis format.
    :param parents: A list of parents for each joint j, i.e. parent[j] is the parent of joint j.
    :param output_format: 'aa' or 'rotmat'.
    :param input_format: 'aa' or 'rotmat'
    :return: The global joint angles as a tensor of shape (N, N_JOINTS*DOF).
    """
    assert output_format in ["aa", "rotmat"]
    assert input_format in ["aa", "rotmat"]
    dof = 3 if input_format == "aa" else 9
    n_joints = poses.shape[-1] // dof
    if input_format == "aa":
        local_oris = aa2rot(poses.reshape((-1, 3)))
    else:
        local_oris = poses
    local_oris = local_oris.reshape((-1, n_joints, 3, 3))
    global_oris = torch.zeros_like(local_oris)

    for j in range(n_joints):
        if parents[j] < 0:
            # root rotation
            global_oris[..., j, :, :] = local_oris[..., j, :, :]
        else:
            parent_rot = global_oris[..., parents[j], :, :]
            local_rot = local_oris[..., j, :, :]
            global_oris[..., j, :, :] = torch.matmul(parent_rot, local_rot)

    if output_format == "aa":
        global_oris = rot2aa(global_oris.reshape((-1, 3, 3)))
        res = global_oris.reshape((-1, n_joints * 3))
    else:
        res = global_oris.reshape((-1, n_joints * 3 * 3))
    return res



import os

import numpy as np
import trimesh
from utils import RESOURCE_DIR, reference, requires_smpl, viewer

from aitviewer.configuration import CONFIG as C
from aitviewer.models.smpl import SMPLLayer
from aitviewer.renderables.lines import Lines
from aitviewer.renderables.meshes import Meshes
from aitviewer.renderables.rigid_bodies import RigidBodies
from aitviewer.renderables.smpl import SMPLSequence
from aitviewer.renderables.spheres import Spheres
from aitviewer.utils.so3 import aa2rot_numpy as aa2rot
from aitviewer.viewer import Viewer


@reference()
def test_renderables(viewer: Viewer):
    grid_xz = np.mgrid[-1.5:1.5:0.3, -1.5:1.5:0.3]
    n_lines = grid_xz.shape[1] * grid_xz.shape[2]

    xz_coords = np.reshape(grid_xz, (2, -1)).T
    line_starts = np.concatenate([xz_coords[:, 0:1], np.zeros((n_lines, 1)), xz_coords[:, 1:2]], axis=-1)
    line_ends = line_starts.copy()
    line_ends[:, 1] = 1.0
    line_strip = np.zeros((2 * n_lines, 3))
    line_strip[::2] = line_starts
    line_strip[1::2] = line_ends
    line_renderable = Lines(line_strip, r_base=0.03, mode="lines")

    line_dirs = line_ends - line_starts
    sphere_positions = line_ends + 0.1 * (line_dirs / np.linalg.norm(line_dirs, axis=-1, keepdims=True))
    spheres = Spheres(sphere_positions, radius=0.05, color=(1.0, 0.0, 1.0, 1.0))

    rb_positions = line_ends + 0.4 * (line_dirs / np.linalg.norm(line_dirs, axis=-1, keepdims=True))
    angles = np.arange(0.0, 2 * np.pi, step=2 * np.pi / n_lines)[:, None]
    axes = np.zeros((n_lines, 3))
    axes[:, 2] = 1.0
    rb_orientations = aa2rot(angles * axes)
    rbs = RigidBodies(rb_positions, rb_orientations)

    viewer.scene.add(line_renderable, spheres, rbs)
    viewer.scene.camera.position = np.array([2.0, 2.0, 2.0])
    viewer.scene.camera.target = np.array([0.0, 0.5, 0.0])


@reference()
def test_obj(viewer: Viewer):
    cube = trimesh.load(os.path.join(RESOURCE_DIR, "cube.obj"), process=False)
    cube_mesh = Meshes(
        cube.vertices,
        cube.faces,
        name="Cube",
        position=[-7.0, 0.0, 0.0],
        flat_shading=True,
    )

    planet = trimesh.load(os.path.join(RESOURCE_DIR, "planet/planet.obj"), process=False)
    texture_image = os.path.join(RESOURCE_DIR, "planet/mars.png")
    planet_mesh = Meshes(
        planet.vertices,
        planet.faces,
        planet.vertex_normals,
        uv_coords=planet.visual.uv,
        path_to_texture=texture_image,
        position=[7.0, 0.0, 0.0],
    )

    drill = trimesh.load(os.path.join(RESOURCE_DIR, "drill/drill.obj"), process=False)
    texture_image = os.path.join(RESOURCE_DIR, "drill/drill_uv.png")
    drill_mesh = Meshes(
        drill.vertices,
        drill.faces,
        drill.vertex_normals,
        uv_coords=drill.visual.uv,
        path_to_texture=texture_image,
        scale=50.0,
        color=(1, 1, 1, 0.5),
    )

    viewer.scene.camera.position = (0, 5, 15)
    viewer.scene.camera.target = (0.15, 0.1, 0.33)
    viewer.auto_set_camera_target = False
    viewer.scene.add(planet_mesh, drill_mesh, cube_mesh)


@reference()
@requires_smpl
def test_smplh(viewer: Viewer):
    smplh_male = SMPLSequence.t_pose(
        SMPLLayer(model_type="smplh", gender="male", device=C.device),
        name="SMPL",
        position=np.array((-1, 0.0, 0.0)),
    )
    smplh_female = SMPLSequence.t_pose(
        SMPLLayer(model_type="smplh", gender="female", device=C.device),
        name="SMPL",
        position=np.array((1, 0.0, 0.0)),
    )
    viewer.scene.camera.position = np.array([0.0, 0.5, 3.5])
    viewer.scene.add(smplh_male, smplh_female)


@reference()
@requires_smpl
def test_smpl(viewer: Viewer):
    smpl_male = SMPLSequence.t_pose(
        SMPLLayer(model_type="smpl", gender="male", device=C.device),
        name="SMPL",
        position=np.array((-1.5, 0, 0)),
    )
    smpl_neutral = SMPLSequence.t_pose(
        SMPLLayer(model_type="smpl", gender="neutral", device=C.device),
        name="SMPL",
        position=np.array((0, 0, 0)),
    )
    smpl_female = SMPLSequence.t_pose(
        SMPLLayer(model_type="smpl", gender="female", device=C.device),
        name="SMPL",
        position=np.array((1.5, 0, 0)),
    )
    viewer.scene.camera.position = np.array([0.0, 0.5, 3.5])
    viewer.scene.add(smpl_male, smpl_female, smpl_neutral)


@reference()
@requires_smpl
def test_smplx(viewer: Viewer):
    smplx_male = SMPLSequence.t_pose(
        SMPLLayer(model_type="smplx", gender="male", device=C.device),
        name="SMPL",
        position=np.array((-1.5, 0, 0)),
    )
    smplx_neutral = SMPLSequence.t_pose(
        SMPLLayer(model_type="smplx", gender="neutral", device=C.device),
        name="SMPL",
        position=np.array((0, 0, 0)),
    )
    smplx_female = SMPLSequence.t_pose(
        SMPLLayer(model_type="smplx", gender="female", device=C.device),
        name="SMPL",
        position=np.array((1.5, 0, 0)),
    )
    viewer.scene.camera.position = np.array([0.0, 0.5, 3.5])
    viewer.scene.add(smplx_male, smplx_female, smplx_neutral)


@reference(count=3)
@requires_smpl
def test_amass(viewer: Viewer):
    c = (149 / 255, 85 / 255, 149 / 255, 0.5)
    seq_amass = SMPLSequence.from_amass(
        npz_data_path=os.path.join(C.datasets.amass, "ACCAD/Female1Running_c3d/C2 - Run to stand_poses.npz"),
        fps_out=60.0,
        name="AMASS Running",
        color=c,
        show_joint_angles=True,
        log=False,
    )
    viewer.scene.camera.position = np.array([-3.3, 1.4, 0.2])
    viewer.scene.camera.target = np.array([-2.8, 1.0, -1.6])
    viewer.scene.add(seq_amass)



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev, Dario Mylonopoulos

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np
import scipy.ndimage

from aitviewer.scene.node import Node


def line(start, end, num):
    """
    Returns an array of positions on a straight line (num, 3).
    :param start: start position, np array-like of shape (3).
    :param end: end position, np array-like of shape (3).
    :param num: number of positions in the returned array.
    """
    return np.linspace(np.array(start), np.array(end), num=num)


def circle(center, radius, num, start_angle=0.0, end_angle=360.0):
    """
    Returns an array of positions on a circle (num_frames, 3).
    :param center: center position of the circle, np array-like of shape (3).
    :param radius: radius of the circle, float.
    :param num: number of positions in the returned array.
    :param start_angle: starting angle on the circle in degrees.
    :param end_angle: ending angle on the circle in degrees.
    """
    angles = np.linspace(np.radians(start_angle), np.radians(end_angle), num=num)
    c = np.column_stack((np.cos(angles) * radius, np.zeros(angles.shape), np.sin(angles) * radius))
    return c + center


def lock_to_node(node: Node, relative_position, smooth_sigma=None):
    """
    Returns a tuple (positions, targets) of two arrays of shape (N, 3) each where N is the
    number of frames of the node. Each target is computed as the center position of
    the node at each frame and each position is at a constant offset from the target by the amount
    specified in the parameter 'relative_position'. The two arrays can be smoothed
    by passing a value greater than 0 to smooth_sigma.

    :param node: the Node to follow.
    :param relative_position: A position added to the node position to compute the positions array.
    :param smooth_sigma: if not None and greater than 0 the position and target arrays are smoothed.
      with a 1D gaussian kernel of standard deviation equal to this value.
    """
    assert isinstance(node, Node), "Node parameter must be a node"

    relative_position = np.array(relative_position)
    old_current_frame_id = node.current_frame_id

    positions = np.zeros((node.n_frames, 3))
    targets = np.zeros((node.n_frames, 3))
    for i in range(node.n_frames):
        node.current_frame_id = i
        target = node.current_center
        position = target + relative_position
        positions[i] = position
        targets[i] = target

    node.current_frame_id = old_current_frame_id

    if smooth_sigma is not None and smooth_sigma > 0:
        positions = scipy.ndimage.gaussian_filter1d(positions, smooth_sigma, axis=0, mode="nearest")
        targets = scipy.ndimage.gaussian_filter1d(targets, smooth_sigma, axis=0, mode="nearest")

    return positions, targets



import os
import shutil

import numpy as np
from PIL import Image
from pxr import Sdf, UsdShade


def _get_texture_paths(path, name, directory):
    usd_path = os.path.join("textures", name)
    fs_path = os.path.join(directory, usd_path)
    os.makedirs(fs_path, exist_ok=True)
    return os.path.join(fs_path, os.path.basename(path)), os.path.join(usd_path, os.path.basename(path))


def copy_texture(path, name, directory):
    fs_path, usd_path = _get_texture_paths(path, name, directory)
    shutil.copy(path, fs_path)
    return usd_path


def save_image_as_texture(img, img_name, name, directory):
    fs_path, usd_path = _get_texture_paths(img_name, name, directory)
    if isinstance(img, np.ndarray):
        img = Image.fromarray(img)
    img.save(fs_path)
    return usd_path


def add_texture(stage, mesh, usd_path, texture_path):
    # Material.
    mat_path = usd_path + "/material"
    material = UsdShade.Material.Define(stage, mat_path)
    input = material.CreateInput("frame:stPrimvarName", Sdf.ValueTypeNames.Token)
    input.Set("st")

    # Shader.
    shader = UsdShade.Shader.Define(stage, mat_path + "/shader")
    shader.CreateIdAttr("UsdPreviewSurface")

    # Connect the material to the shader.
    material.CreateSurfaceOutput().ConnectToSource(shader.ConnectableAPI(), "surface")

    # Create a texture coordinates reader and connect it to the material input.
    reader = UsdShade.Shader.Define(stage, mat_path + "/stReader")
    reader.CreateIdAttr("UsdPrimvarReader_float2")
    reader.CreateInput("varname", Sdf.ValueTypeNames.Token).ConnectToSource(input)

    # Create a texture.
    diffuse = UsdShade.Shader.Define(stage, mat_path + "/diffuseTexture")
    diffuse.CreateIdAttr("UsdUVTexture")
    diffuse.CreateInput("file", Sdf.ValueTypeNames.Asset).Set(texture_path)
    diffuse.CreateInput("st", Sdf.ValueTypeNames.Float2).ConnectToSource(reader.ConnectableAPI(), "result")
    diffuse.CreateOutput("rgb", Sdf.ValueTypeNames.Float3)
    shader.CreateInput("diffuseColor", Sdf.ValueTypeNames.Color3f).ConnectToSource(diffuse.ConnectableAPI(), "rgb")

    # Bind the Material to the mesh.
    mesh.GetPrim().ApplyAPI(UsdShade.MaterialBindingAPI)
    UsdShade.MaterialBindingAPI(mesh).Bind(material)



from ..message import Message
from ..node import RemoteNode


class RemoteRigidBodies(RemoteNode):
    MESSAGE_TYPE = Message.RIGID_BODIES

    def __init__(self, viewer, rb_pos, rb_ori, **kwargs):
        """
        This initializer takes a RemoteViewer object and all other arguments are forwarded
        to the RigidBodies constructor on the remote Viewer.
        See the RigidBodies class for more information about parameters.

        :param rb_pos: A np array of shape (N, R, 3) or (R, 3) rigid-body positions.
        :param rb_ori: A np array of shape (N, R, 3, 3) or (R, 3, 3) rigid-body orientations.
        """
        super().__init__(
            viewer,
            rb_pos=rb_pos,
            rb_ori=rb_ori,
            **kwargs,
        )

    def add_frames(self, rb_pos, rb_ori):
        """
        Add frames to the remote RigidBodies node by adding new positions and orientations.

        :param rb_pos: A np array of shape (N, R, 3) or (R, 3) rigid-body positions.
        :param rb_ori: A np array of shape (N, R, 3, 3) or (R, 3, 3) rigid-body orientations.
        """
        return super().add_frames(rb_pos=rb_pos, rb_ori=rb_ori)

    def update_frames(self, rb_pos, rb_ori, frames):
        """
        Update frames of the remote RigidBodies node by updating the positions and orientations.

        :param rb_pos: A np array of shape (N, R, 3) or (R, 3) rigid-body positions.
        :param rb_ori: A np array of shape (N, R, 3, 3) or (R, 3, 3) rigid-body orientations.
        :param frames: a list of integer frame indices of size N or a single integer frame index.
        """
        return super().update_frames(rb_pos=rb_pos, rb_ori=rb_ori, frames=frames)



"""
Copyright (C) 2023  ETH Zurich, Manuel Kaufmann, Velko Vechev, Dario Mylonopoulos

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import collections
import os

import torch
import torch.nn.functional as F

try:
    from supr.config import cfg
    from supr.pytorch.supr import SUPR
except Exception as e:
    raise ImportError(f"Cannot import SUPR. Please run `pip install git+https://github.com/ahmedosman/SUPR.git`\n{e}")

from aitviewer.configuration import CONFIG as C
from aitviewer.utils.so3 import aa2rot_torch as aa2rot
from aitviewer.utils.so3 import rot2aa_torch as rot2aa


class SUPRLayer(SUPR):
    """Wraps the publicly available SUPR model to match SMPLX model interface"""

    def __init__(self, gender="male", num_betas=10, constrained=False, device=None, dtype=None):
        """
        Initializer.
        :param gender: Which gender to load.
        :param num_betas: Number of shape components.
        :param device: CPU or GPU.
        :param dtype: The pytorch floating point data type.
        """
        # Configure SUPR model before initializing
        cfg.data_type = dtype if dtype is not None else C.f_precision
        path_model = os.path.join(C.supr_models, f'supr_{gender}{"_constrained" if constrained else ""}.npy')
        super(SUPRLayer, self).__init__(path_model, num_betas=num_betas)

        self.device = device if device is not None else C.device
        self.model_type = "supr"
        self._parents = None
        self._children = None

    @property
    def parents(self):
        """Return how the joints are connected in the kinematic chain where parents[i, 0] is the parent of
        joint parents[i, 1]."""
        if self._parents is None:
            self._parents = self.kintree_table.transpose(0, 1).cpu().numpy()
        return self._parents

    @property
    def joint_children(self):
        """Return the children of each joint in the kinematic chain."""
        if self._children is None:
            self._children = collections.defaultdict(list)
            for bone in self.parents:
                if bone[0] != -1:
                    self._children[bone[0]].append(bone[1])
        return self._children

    def skeletons(self):
        """Return how the joints are connected in the kinematic chain where skeleton[0, i] is the parent of
        joint skeleton[1, i]."""
        kintree_table = self.kintree_table
        kintree_table[:, 0] = -1
        return {
            "all": kintree_table,
            "body": kintree_table[:, : self.n_joints_body + 1],
        }

    @property
    def n_joints_body(self):
        return self.parent.shape[0]

    @property
    def n_joints_total(self):
        return self.n_joints_body + 1

    def forward(self, poses_body, betas=None, poses_root=None, trans=None, normalize_root=False):
        """
        forwards the model
        :param poses_body: Pose parameters.
        :param poses_root: Pose parameters for the root joint.
        :param beta: Beta parameters.
        :param trans: Root translation.
        :param normalize_root: Makes poses relative to the root joint (useful for globally rotated captures).
        :return: Deformed surface vertices, transformed joints
        """
        poses, betas, trans = self.preprocess(poses_body, betas, poses_root, trans, normalize_root)

        # SUPR repo currently hardcodes floats.
        v = super().forward(pose=poses.float(), betas=betas.float(), trans=trans.float())
        J = v.J_transformed
        return v, J

    def preprocess(self, poses_body, betas=None, poses_root=None, trans=None, normalize_root=False):
        batch_size = poses_body.shape[0]

        if poses_root is None:
            poses_root = torch.zeros([batch_size, 3]).to(dtype=poses_body.dtype, device=self.device)
        if trans is None:
            # If we don't supply the root translation explicitly, it falls back to using self.bm.trans
            # which might not be zero since it is a trainable param that can get updated.
            trans = torch.zeros([batch_size, 3]).to(dtype=poses_body.dtype, device=self.device)

        if normalize_root:
            # Make everything relative to the first root orientation.
            root_ori = aa2rot(poses_root)
            first_root_ori = torch.inverse(root_ori[0:1])
            root_ori = torch.matmul(first_root_ori, root_ori)
            poses_root = rot2aa(root_ori)
            trans = torch.matmul(first_root_ori.unsqueeze(0), trans.unsqueeze(-1)).squeeze()
            trans = trans - trans[0:1]

        poses = torch.cat((poses_root, poses_body), dim=1)

        if betas is None:
            betas = torch.zeros([batch_size, self.num_betas]).to(dtype=poses_body.dtype, device=self.device)

        # Batch shapes if they don't match batch dimension.
        if betas.shape[0] != batch_size:
            betas = betas.repeat(batch_size, 1)

        # Lower bound betas
        if betas.shape[1] < self.num_betas:
            betas = torch.nn.functional.pad(betas, [0, self.num_betas - betas.shape[1]])

        # Upper bound betas
        betas = betas[:, : self.num_betas]

        return poses, betas, trans



from .message import Message
from .viewer import RemoteViewer

GLOBAL_ID = 0


def next_id():
    global GLOBAL_ID
    GLOBAL_ID += 1
    return GLOBAL_ID


class RemoteNode:
    MESSAGE_TYPE = Message.NODE

    def __init__(self, viewer: RemoteViewer, *args, **kwargs):
        """
        This initializer takes a RemoteViewer object and all other arguments are forwarded
        to the Node constructor on the remote Viewer.
        See the Node class for more information about parameters.

        :param viewer: a RemoteViewer object that will be used to send this node.
        """
        self.viewer = viewer
        self.uid = next_id()
        self._send_msg(self.MESSAGE_TYPE, *args, **kwargs)

    def _send_msg(self, type, *args, **kwargs):
        # Helper function to send a message with the node uid.
        self.viewer.send_message(type, self.uid, *args, **kwargs)

    def add_frames(self, *args, **kwargs):
        """
        Add frames to the remote Node.
        This function should be implemented by subclasses to match
        the relative add_frames() method on the renderable class.
        """
        self._send_msg(Message.ADD_FRAMES, *args, **kwargs)

    def update_frames(self, *args, **kwargs):
        """
        Update frames of the remote Node.
        This function should be implemented by subclasses to match
        the relative add_frames() method on the renderable class.
        """
        self._send_msg(Message.UPDATE_FRAMES, *args, **kwargs)

    def remove_frames(self, frames):
        """
        Remove frames of the remote Node.

        :param frames: a list of integer frame indices of size N or a single integer frame index to remove.
        """
        self._send_msg(Message.REMOVE_FRAMES, frames=frames)

    def delete(self):
        """Delete the remote Node from the scene."""
        self._send_msg(Message.DELETE)



import asyncio
import pickle
import queue
import subprocess
import threading
from typing import Callable

import websockets

from .message import Message, make_message


class RemoteViewer:
    def __init__(self, host="localhost", port=8417, timeout=10, verbose=True):
        """
        Initializer.
        :param host: the IP address of a host to connect to as a string.
        :param port: the TCP port to connect to.
        :param timeout: a timeout in seconds for attempting to connect to the viewer.
        :param verbose: if True print info messages.
        """
        url = f"ws://{host}:{port}"

        if verbose:
            print(f"Connecting to remote viewer at {url}")

        self.timeout = timeout
        self.connected = False

        # Semaphore used to wait for the connection to be setup by the client thread.
        self.semaphore = threading.Semaphore(0)

        # Create a thread for running the websocket client async loop.
        self.loop = asyncio.new_event_loop()
        self.thread = threading.Thread(target=self._entry, args=(url,), daemon=True)
        self.thread.start()

        # Wait for the connection to be setup.
        self.semaphore.acquire()
        if verbose:
            if self.connected:
                print("Connected")
            else:
                print(f"Failed to connect")

        self.process: subprocess.Popen = None

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close_connection()

    @classmethod
    def create_new_process(cls, args=None, **kwargs):
        """
        Open a Viewer in a new process and return a RemoteViewer connected to it.

        :param args: This parameter can be used to specify an argument or
          a list of arguments that is used to create the new process.
          e.g: args = ["path/to/script.py", "arg1", "arg2"] will invoke the following command:
                python path/to/script.py arg1 arg2
        """
        # If host is None create a new viewer in a separate process.
        if args is None:
            popen_args = ["python", "-m", "aitviewer.server"]
        else:
            if isinstance(args, list):
                popen_args = ["python"] + args
            else:
                popen_args = ["python", str(args)]

        # Create the viewer process.
        process = subprocess.Popen(
            popen_args,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

        # Create a remote viewer connected to the child process.
        v = cls(**kwargs)
        v.process = process

        return v

    def _entry(self, url):
        # Entry point of the client thread.

        asyncio.set_event_loop(self.loop)
        self.loop.run_until_complete(self._async_entry(url))

    async def _async_entry(self, url):
        # Async entry point of the client thread.

        # Attempt to connect until 'self.timeout' seconds passed.
        start_time = self.loop.time()
        try:
            while self.loop.time() < start_time + self.timeout:
                try:
                    self.websocket = await websockets.connect(url, max_size=None)
                    self.connected = True
                    break
                except Exception as e:
                    pass
        finally:
            # Release the semaphore to let the main thread continue after
            # attempting to connect. The main thread will read the
            # self.connected variable to know if we succeded at connecting.
            self.semaphore.release()

        # Exit the client thread if we failed to connect.
        if not self.connected:
            return

        # Create a queue for incoming messages to the main thread.
        self.recv_queue = queue.Queue()

        # Message loop.
        try:
            # This loop is exited whenever the connection is dropped
            # which causes and exception to be raised.
            async for message in self.websocket:
                data = pickle.loads(message)
                # Equeue data for the main thread to process.
                self.recv_queue.put_nowait(data)
        except Exception as e:
            print(f"Message loop exception: {e}")

        # Mark the connection as closed.
        self.connected = False

    def get_message(self, block=True):
        """
        Returns the next message received by the remote viewer.

        :param block: if True this function blocks until a message is received, otherwise it returns immediately.

        :return: if block is True returns the next message or None if the connection has been closed.
                 if block is False returns the next message or None if there are no messages.
        """
        if self.connected:
            if block:
                while self.connected:
                    try:
                        return self.recv_queue.get(timeout=0.1)
                    except queue.Empty:
                        pass
            else:
                if not self.recv_queue.empty():
                    return self.recv_queue.get_nowait()

        return None

    def process_messages(self, handler: Callable[["RemoteViewer", object], None], block=True):
        """
        Processes messages in a loop calling 'handler' for each message.

        :param block: if True this function blocks until the connection is closed, otherwise it returns
            after all messages received so far have been processed.

        :return: if block is True always returns False when the connection has been closed.
                 if block is False returns True if the connection is still open or False if the connection
                 has been closed.
        """
        while True:
            msg = self.get_message(block)
            if msg is None:
                if block:
                    return False
                else:
                    return self.connected
            handler(self, msg)

    async def _async_send(self, data):
        await self.websocket.send(data)

    def send(self, data):
        try:
            if self.connected:
                # Send a message by adding a send coroutine to the thread's loop and wait for it to complete.
                asyncio.run_coroutine_threadsafe(self._async_send(data), self.loop).result()
        except Exception as e:
            print(f"Send exception: {e}")

    def send_message(self, type, uid=None, *args, **kwargs):
        """
        Send a message to the viewer. See Viewer.process_message()
        for information about how these parameters are interpreted
        by the viewer.
        """
        msg = make_message(type, uid, args, kwargs)
        data = pickle.dumps(msg)
        self.send(data)

    def set_frame(self, frame: int):
        """
        Set the current active frame of the remote viewer.

        :param frame: an integer representing the id of the frame.
        """
        self.send_message(Message.SET_FRAME, None, frame)

    def next_frame(self):
        """Set the current active frame of the remote viewer to the next frame"""
        self.send_message(Message.NEXT_FRAME)

    def previous_frame(self):
        """Set the current active frame of the remote viewer to the previous frame"""
        self.send_message(Message.PREVIOUS_FRAME)

    async def _async_close(self):
        await self.websocket.close()

    def close_connection(self):
        """Close the connection with the remote viewer."""
        if self.connected:
            asyncio.run_coroutine_threadsafe(self._async_close(), self.loop).result()

            # Wait for the client thread to exit.
            self.thread.join()

    def wait_process(self, print_viewer_output=True):
        """
        If the viewer was created locally in a separate process wait for it
        to exit and optionally print the standard output of the remote viewer.

        :param print_viewer_output: if True print the output of the remote viewer.
        """
        self.close_connection()
        if self.process is not None:
            self.process.wait()
            if print_viewer_output:
                print("\nRemote viewer output:")
                print(self.process.stdout.read().decode())



import struct
from typing import Dict, List, Tuple

import moderngl
import numpy as np
from moderngl_window import BaseWindow, geometry
from moderngl_window.opengl.vao import VAO
from PIL import Image

from aitviewer.scene.camera import CameraInterface, ViewerCamera
from aitviewer.scene.node import Node
from aitviewer.scene.scene import Scene
from aitviewer.shaders import load_program


class Viewport:
    def __init__(self, extents: Tuple[int, int, int, int], camera: CameraInterface):
        self.extents = extents
        self._camera = camera
        self._using_temp_camera = not isinstance(camera, ViewerCamera)

    @property
    def camera(self):
        return self._camera

    @camera.setter
    def camera(self, camera: CameraInterface):
        self._camera = camera
        self._using_temp_camera = not isinstance(camera, ViewerCamera)

    @property
    def using_temp_camera(self):
        return self._using_temp_camera

    def contains(self, x: int, y: int):
        e = self.extents
        return x >= e[0] and x < e[0] + e[2] and y >= e[1] and y < e[1] + e[3]

    def set_temp_camera(self, camera: CameraInterface):
        self.camera = camera
        self._using_temp_camera = True

    def reset_camera(self):
        if self._using_temp_camera:
            self._using_temp_camera = False

            fwd = self.camera.forward
            pos = self.camera.position

            self.camera = ViewerCamera(45)
            self.camera.position = np.copy(pos)
            self.camera.target = pos + fwd * 3
            self.camera.update_matrices(*self.extents[2:])


class Renderer:
    def __init__(self, ctx: moderngl.Context, wnd: BaseWindow, window_type: str):
        self.ctx = ctx
        self.wnd = wnd
        self.window_type = window_type

        # Shaders for mesh mouse intersection.
        self.frag_pick_prog = load_program("fragment_picking/frag_pick.glsl")
        self.frag_pick_prog["position_texture"].value = 0  # Read from texture channel 0
        self.frag_pick_prog["obj_info_texture"].value = 1  # Read from texture channel 0
        self.picker_output = self.ctx.buffer(reserve=6 * 4)  # 3 floats, 3 ints
        self.picker_vao = VAO(mode=moderngl.POINTS)

        # Shaders for drawing outlines.
        self.outline_draw_prog = load_program("outline/outline_draw.glsl")
        self.outline_quad = geometry.quad_2d(size=(2.0, 2.0), pos=(0.0, 0.0))

        # Create framebuffers
        self.offscreen_p_depth = None
        self.offscreen_p_viewpos = None
        self.offscreen_p_tri_id = None
        self.offscreen_p = None
        self.outline_texture = None
        self.outline_framebuffer = None
        self.headless_fbo_color = None
        self.headless_fbo_depth = None
        self.headless_fbo = None
        self.create_framebuffers()

        # Debug
        self.vis_prog = load_program("visualize.glsl")
        self.vis_quad = geometry.quad_2d(size=(0.9, 0.9), pos=(0.5, 0.5))

    def clear(self, scene: Scene, **kwargs):
        """Clear the window framebuffer and the fragmap framebuffer."""
        # Clear picking buffer.
        self.offscreen_p.clear()

        # Clear background and make sure only the flags we want are enabled.
        if kwargs["transparent_background"]:
            self.wnd.clear(0, 0, 0, 0)
        else:
            if scene.light_mode == "dark":
                self.wnd.clear(0.1, 0.1, 0.1, 1.0)
            else:
                self.wnd.clear(*scene.background_color)

    def create_framebuffers(self):
        """
        Create all framebuffers which depend on the window size.
        This is called once at startup and every time the window is resized.
        """

        # Release framebuffers if they already exist.
        def safe_release(b):
            if b is not None:
                b.release()

        safe_release(self.offscreen_p_depth)
        safe_release(self.offscreen_p_viewpos)
        safe_release(self.offscreen_p_tri_id)
        safe_release(self.offscreen_p)
        safe_release(self.outline_texture)
        safe_release(self.outline_framebuffer)

        # Mesh mouse intersection
        self.offscreen_p_depth = self.ctx.depth_texture(self.wnd.buffer_size)
        self.offscreen_p_viewpos = self.ctx.texture(self.wnd.buffer_size, 4, dtype="f4")
        self.offscreen_p_tri_id = self.ctx.texture(self.wnd.buffer_size, 4, dtype="f4")
        self.offscreen_p = self.ctx.framebuffer(
            color_attachments=[self.offscreen_p_viewpos, self.offscreen_p_tri_id],
            depth_attachment=self.offscreen_p_depth,
        )
        self.offscreen_p_tri_id.filter = (moderngl.NEAREST, moderngl.NEAREST)

        # Outline rendering
        self.outline_texture = self.ctx.texture(self.wnd.buffer_size, 1, dtype="f4")
        self.outline_framebuffer = self.ctx.framebuffer(color_attachments=[self.outline_texture])

        # If in headlesss mode we create a framebuffer without multisampling that we can use
        # to resolve the default framebuffer before reading.
        if self.window_type == "headless":
            safe_release(self.headless_fbo_color)
            safe_release(self.headless_fbo_depth)
            safe_release(self.headless_fbo)
            self.headless_fbo_color = self.ctx.texture(self.wnd.buffer_size, 4)
            self.headless_fbo_depth = self.ctx.depth_texture(self.wnd.buffer_size)
            self.headless_fbo = self.ctx.framebuffer(self.headless_fbo_color, self.headless_fbo_depth)

    def render_shadowmap(self, scene: Scene, **kwargs):
        """A pass to render the shadow map, i.e. render the entire scene once from the view of the light."""
        if kwargs["shadows_enabled"]:
            self.ctx.enable_only(moderngl.DEPTH_TEST)
            rs = scene.collect_nodes()

            for light in scene.lights:
                if light.shadow_enabled:
                    light.use(self.ctx)
                    light_matrix = light.mvp()
                    for r in rs:
                        r.render_shadowmap(light_matrix)

    def render_fragmap(self, scene: Scene, camera: CameraInterface, viewport):
        """A pass to render the fragment picking map, i.e. render the scene with world coords as colors."""
        self.ctx.enable_only(moderngl.DEPTH_TEST)
        self.offscreen_p.use()
        self.offscreen_p.viewport = viewport

        rs = scene.collect_nodes()
        for r in rs:
            r.render_fragmap(self.ctx, camera)

    def render_scene(self, scene: Scene, camera: CameraInterface, viewport, **kwargs):
        """Render the current scene to the framebuffer without time accounting and GUI elements."""
        # Bind framebuffer.
        self.wnd.use()

        # Setup viewport.
        self.wnd.fbo.viewport = viewport

        # Configure OpenGL context.
        self.ctx.enable_only(moderngl.DEPTH_TEST | moderngl.BLEND | moderngl.CULL_FACE)
        self.ctx.cull_face = "back"
        self.ctx.blend_func = (
            moderngl.SRC_ALPHA,
            moderngl.ONE_MINUS_SRC_ALPHA,
            moderngl.ONE,
            moderngl.ONE,
        )

        # Render scene.
        scene.render(
            camera=camera,
            window_size=kwargs["window_size"],
            lights=scene.lights,
            shadows_enabled=kwargs["shadows_enabled"],
            ambient_strength=scene.ambient_strength,  # ???
            fbo=self.wnd.fbo,
        )

    def render_outline(self, nodes, color, camera: CameraInterface, viewport):
        """A pass to render outlines of an object."""
        # Prepare the outline buffer, all objects rendered to this buffer will be outlined.
        self.outline_framebuffer.clear()
        self.outline_framebuffer.use()
        self.outline_framebuffer.viewport = viewport

        # Render outline of the nodes with outlining enabled, this potentially also renders their children.
        for n in nodes:
            n.render_outline(self.ctx, camera)

        # Render the outline effect to the window.
        self.wnd.use()
        self.wnd.fbo.depth_mask = False
        self.wnd.fbo.viewport = viewport
        self.ctx.enable_only(moderngl.NOTHING)
        self.outline_texture.use(0)
        self.outline_draw_prog["outline"] = 0
        self.outline_draw_prog["outline_color"] = color
        self.outline_quad.render(self.outline_draw_prog)
        self.wnd.fbo.depth_mask = True

    def render_viewport(self, scene: Scene, camera: CameraInterface, viewport: Viewport, **kwargs):
        """Render the scene for a viewport with the given camera."""
        # Parameters.
        export = kwargs["export"]
        outline_color = kwargs["outline_color"]
        light_outline_color = kwargs["light_outline_color"]
        selected_outline_color = kwargs["selected_outline_color"]

        # Update camera to viewport.
        width, height = viewport[2:]
        camera.update_matrices(width, height)

        # Disable some renderables for exporting.
        renderables_to_enable: List[Node] = []
        if export:

            def disable_for_export(r: Node):
                if r.enabled:
                    r.enabled = False
                    renderables_to_enable.append(r)

            # Disable lights.
            for l in scene.lights:
                disable_for_export(l.mesh)

            # Disable camera target.
            disable_for_export(scene.camera_target)

        self.render_fragmap(scene, camera, viewport)
        self.render_scene(scene, camera, viewport, **kwargs)
        self.render_outline(
            [n for n in scene.collect_nodes() if n.draw_outline],
            outline_color,
            camera,
            viewport,
        )

        # Re-enable renderables that were disabled for exporting.
        if export:
            for r in renderables_to_enable:
                r.enabled = True

        if not export:
            self.render_outline([l for l in scene.lights if l.enabled], light_outline_color, camera, viewport)

            # If the selected object is a Node render its outline.
            if isinstance(scene.selected_object, Node):
                self.render_outline([scene.selected_object], selected_outline_color, camera, viewport)

    def resize(self, _width: int, _height: int):
        """Resize rendering resources on window resize."""
        self.create_framebuffers()

    def render(self, scene: Scene, **kwargs):
        """Main rendering function."""
        viewports: List[Viewport] = kwargs["viewports"]

        # Viewport 0 always matches the scene camera for backwards compatitibility.
        viewports[0].camera = scene.camera

        # Render to shadowmap, this is can be done once for all viewports.
        self.render_shadowmap(scene, **kwargs)
        self.clear(scene, **kwargs)

        for v in viewports:
            self.render_viewport(scene, v.camera, v.extents, **kwargs)
        self.wnd.fbo.viewport = [0, 0, self.wnd.buffer_size[0], self.wnd.buffer_size[1]]

        if not kwargs["export"]:
            # If visualize is True draw a texture with the object id to the screen for debugging.
            if kwargs["visualize"]:
                self.ctx.enable_only(moderngl.NOTHING)
                self.offscreen_p_tri_id.use(location=0)
                self.vis_prog["hash_color"] = True
                self.vis_quad.render(self.vis_prog)

    def read_fragmap_at_pixel(self, x: int, y: int) -> Tuple[np.ndarray, int, int, int]:
        """Given an x/y screen coordinate, get the intersected object, triangle id, and xyz point in camera space."""

        # Fragment picker uses already encoded position/object/triangle in the frag_pos program textures
        self.frag_pick_prog["texel_pos"].value = (x, y)
        self.offscreen_p_viewpos.use(location=0)
        self.offscreen_p_tri_id.use(location=1)
        self.picker_vao.transform(self.frag_pick_prog, self.picker_output, vertices=1)
        x, y, z, obj_id, tri_id, instance_id = struct.unpack("3f3i", self.picker_output.read())
        return np.array((x, y, z)), obj_id, tri_id, instance_id

    def get_current_frame_as_image(self, alpha=False):
        """Return the FBO content as a PIL image."""
        if alpha:
            fmt = "RGBA"
            components = 4
        else:
            fmt = "RGB"
            components = 3

        # If in headless mode we first resolve the multisampled framebuffer into
        # a non multisampled one and read from that instead.
        if self.window_type == "headless":
            self.ctx.copy_framebuffer(self.headless_fbo, self.wnd.fbo)
            fbo = self.headless_fbo
        else:
            fbo = self.wnd.fbo

        width = self.wnd.fbo.viewport[2] - self.wnd.fbo.viewport[0]
        height = self.wnd.fbo.viewport[3] - self.wnd.fbo.viewport[1]
        image = Image.frombytes(
            fmt,
            (width, height),
            fbo.read(viewport=self.wnd.fbo.viewport, alignment=1, components=components),
        )
        if width != self.wnd.size[0] or height != self.wnd.size[1]:
            image = image.resize(self.wnd.size, Image.NEAREST)

        return image.transpose(Image.FLIP_TOP_BOTTOM)

    def get_current_depth_image(self, camera: CameraInterface):
        """
        Return the depth buffer as a 'F' PIL image.
        Depth is stored as the z coordinate in eye (view) space.
        Therefore values in the depth image represent the distance from the pixel to
        the plane passing through the camera and orthogonal to the view direction.
        Values are between the near and far plane distances of the camera used for rendering,
        everything outside this range is clipped by OpenGL.
        """

        # If in headless mode we first resolve the multisampled framebuffer into
        # a non multisampled one and read from that instead.
        if self.window_type == "headless":
            self.ctx.copy_framebuffer(self.headless_fbo, self.wnd.fbo)
            fbo = self.headless_fbo
        else:
            fbo = self.wnd.fbo

        width = self.wnd.fbo.viewport[2] - self.wnd.fbo.viewport[0]
        height = self.wnd.fbo.viewport[3] - self.wnd.fbo.viewport[1]

        # Get depth image from depth buffer.
        depth = Image.frombytes(
            "F",
            (width, height),
            fbo.read(viewport=self.wnd.fbo.viewport, alignment=1, attachment=-1, dtype="f4"),
        )

        if width != self.wnd.size[0] or height != self.wnd.size[1]:
            depth = depth.resize(self.wnd.size, Image.NEAREST)

        # Convert from [0, 1] range to [-1, 1] range.
        # This is necessary because our projection matrix computes NDC
        # from [-1, 1], but depth is then stored normalized from 0 to 1.
        depth = np.array(depth) * 2.0 - 1.0

        # Extract projection matrix parameters used for mapping Z coordinates.
        P = camera.get_projection_matrix()
        a, b = P[2, 2], P[2, 3]

        # Linearize depth values. This converts from [-1, 1] range to the
        # view space Z coordinate value, with positive z in front of the camera.
        z = b / (a + depth)
        return Image.fromarray(z, mode="F").transpose(Image.FLIP_TOP_BOTTOM)

    def get_current_mask_ids(self, id_map: Dict[int, int] = None):
        """
        Return a mask as a numpy array of shape (height, width) and type np.uint32.
        Each element in the array is the UID of the node covering that pixel (can be accessed from a node with 'node.uid')
        or zero if not covered.

        :param id_map:
            if not None the UIDs in the mask are mapped using this dictionary to the specified ID.
            The final mask only contains the IDs specified in this mapping and zeros everywhere else.
        """
        width = self.wnd.fbo.viewport[2] - self.wnd.fbo.viewport[0]
        height = self.wnd.fbo.viewport[3] - self.wnd.fbo.viewport[1]

        # Get object ids as floating point numbers from the first channel of
        # the first attachment of the picking framebuffer.
        id = Image.frombytes(
            "F",
            (width, height),
            self.offscreen_p.read(
                viewport=self.wnd.fbo.viewport,
                alignment=1,
                components=1,
                attachment=1,
                dtype="f4",
            ),
        )

        if width != self.wnd.size[0] or height != self.wnd.size[1]:
            id = id.resize(self.wnd.size, Image.NEAREST)

        # Convert the id to integer values.
        id_int = np.asarray(id).astype(dtype=np.uint32)

        # If an id_map is given use this to map the ids.
        if id_map is not None:
            output = np.zeros(id_int.shape, dtype=np.uint32)
            for k, v in id_map.items():
                output[id_int == k] = v
            return output
        else:
            # Copy here because the array constructed from a PIL image is read-only.
            return id_int.copy()

    def get_current_mask_image(self, color_map: Dict[int, Tuple[int, int, int]] = None, id_map: Dict[int, int] = None):
        """
        Return a color mask as a 'RGB' PIL image.
        Each object in the mask has a uniform color computed from the Node UID (can be accessed from a node with 'node.uid').

        :param color_map:
            if not None specifies the color to use for a given Node UID as a tuple (R, G, B) of integer values from 0 to 255.
            If None the color is computed as an hash of the Node UID instead.
        :param id_map:
            if not None the UIDs in the mask are mapped using this dictionary from Node UID to the specified ID.
            This mapping is applied before the color map (or before hashing if the color map is None).
        """

        if color_map is None:
            # If no colormap is given hash the ids mapped with id_map.
            ids = self.get_current_mask_ids(id_map)

            # Hash the ids.
            def hash(h):
                h ^= h >> 16
                h *= 0x85EBCA6B
                h ^= h >> 13
                h *= 0xC2B2AE35
                h ^= h >> 16
                return h

            output = hash(ids)
        else:
            # If a colormap is given use it to map from UID to colors.
            ids = self.get_current_mask_ids()

            def rgb_to_uint(r, g, b):
                return ((b & 0xFF) << 16) | ((g & 0xFF) << 8) | (r & 0xFF)

            output = np.zeros(ids.shape, dtype=np.uint32)
            if id_map is not None:
                # If a id_map is given use it before indexing into the color map
                for k, v in id_map.items():
                    output[ids == k] = rgb_to_uint(*color_map[v])
            else:
                # If no id_map is given use the color_map directly.
                for k, v in color_map.items():
                    output[ids == k] = rgb_to_uint(*v)

        # Convert the hashed ids to an RGBA image and then throw away the alpha channel.
        img = Image.frombytes("RGBA", (ids.shape[1], ids.shape[0]), output.tobytes()).convert("RGB")
        return img.transpose(Image.FLIP_TOP_BOTTOM)



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev, Dario Mylonopoulos

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from moderngl_window.context.pyqt5 import Window
from PyQt5 import QtGui, QtOpenGL, QtWidgets

"""
Window class adapted from  moderngl_window.context.pyqt5.Window.
Only slight changes are made to avoid a crash with Python >= 3.10.
"""


class PyQt5Window(Window):
    def __init__(self, **kwargs):
        super(Window, self).__init__(**kwargs)

        # Specify OpenGL context parameters
        gl = QtOpenGL.QGLFormat()
        gl.setVersion(self.gl_version[0], self.gl_version[1])
        gl.setProfile(QtOpenGL.QGLFormat.CoreProfile)
        gl.setDepthBufferSize(24)
        gl.setDoubleBuffer(True)
        gl.setSwapInterval(1 if self._vsync else 0)
        gl.setAlphaBufferSize(8)

        # Configure multisampling if needed
        if self.samples > 1:
            gl.setSampleBuffers(True)
            gl.setSamples(int(self.samples))

        # We need an application object, but we are bypassing the library's
        # internal event loop to avoid unnecessary work
        self._app = QtWidgets.QApplication([])

        # Create the OpenGL widget
        self._widget = QtOpenGL.QGLWidget(gl)
        self.title = self._title

        # If fullscreen we change the window to match the desktop on the primary screen
        if self.fullscreen:
            rect = QtWidgets.QDesktopWidget().screenGeometry()
            self._width = rect.width()
            self._height = rect.height()
            self._buffer_width = rect.width() * self._widget.devicePixelRatio()
            self._buffer_height = rect.height() * self._widget.devicePixelRatio()

        if self.resizable:
            # Ensure a valid resize policy when window is resizable
            size_policy = QtWidgets.QSizePolicy(
                QtWidgets.QSizePolicy.Expanding,
                QtWidgets.QSizePolicy.Expanding,
            )
            self._widget.setSizePolicy(size_policy)
            self._widget.resize(self.width, self.height)
        else:
            self._widget.setFixedSize(self.width, self.height)

        # Center the window on the screen if in window mode
        if not self.fullscreen:
            center_window_position = (
                # HACK: Here we changed division to integer division,
                # because move requires its arguments to be integers.
                self.position[0] - self.width // 2,
                self.position[1] - self.height // 2,
            )
            self._widget.move(*center_window_position)

        # Needs to be set before show()
        self._widget.resizeGL = self.resize

        self.cursor = self._cursor

        if self.fullscreen:
            self._widget.showFullScreen()
        else:
            self._widget.show()

        # We want mouse position events
        self._widget.setMouseTracking(True)

        # Override event functions in qt
        self._widget.keyPressEvent = self.key_pressed_event
        self._widget.keyReleaseEvent = self.key_release_event
        self._widget.mouseMoveEvent = self.mouse_move_event
        self._widget.mousePressEvent = self.mouse_press_event
        self._widget.mouseReleaseEvent = self.mouse_release_event
        self._widget.wheelEvent = self.mouse_wheel_event
        self._widget.closeEvent = self.close_event
        self._widget.showEvent = self.show_event
        self._widget.hideEvent = self.hide_event

        # Attach to the context
        self.init_mgl_context()

        # Ensure retina and 4k displays get the right viewport
        self._buffer_width = self._width * self._widget.devicePixelRatio()
        self._buffer_height = self._height * self._widget.devicePixelRatio()

        self.set_default_viewport()

    def _set_icon(self, icon_path: str) -> None:
        self._widget.setWindowIcon(QtGui.QIcon(str(icon_path)))



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev, Dario Mylonopoulos

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np

from aitviewer.renderables.lines import Lines
from aitviewer.scene.node import Node


class Arrows(Node):
    """
    Draw an arrow.
    """

    def __init__(
        self,
        origins,
        tips,
        r_base=0.01,
        r_head=0.02,
        p=0.1,
        color=(0.0, 0.0, 0.5, 1.0),
        **kwargs,
    ):
        """
        Initializer.
        :param origins: Set of 3D coordinates of the base of the arrows as a np array of shape (N, B, 3).
        :param tips: Set of 3D coordinates denoting the tip of the arrow (N, T, 3).
        :param r_base: Radius of the base cylinder.
        :param r_head: Radius of the tip cylinder.
        :param p: Percentage of arrow head on the entire length.
        :param color: Color of the line (4-tuple).
        """
        assert origins.shape == tips.shape
        if len(origins.shape) == 2:
            origins = origins[np.newaxis]
            tips = tips[np.newaxis]
        else:
            assert len(origins.shape) == 3
        super(Arrows, self).__init__(n_frames=len(origins), **kwargs)

        self._origins = origins
        self._tips = tips

        # Percentage of arrow head on entire length
        self.p = p

        # Nodes
        self.material.color = color
        self.bases_r = Lines(
            lines=self.get_line_coords(self.origins, self.mid_points),
            mode="lines",
            r_base=r_base,
            color=color,
            cast_shadow=False,
            is_selectable=False,
        )
        self.arrows_r = Lines(
            lines=self.get_line_coords(self.mid_points, self.tips),
            mode="lines",
            r_base=r_head,
            r_tip=0.0,
            color=color,
            cast_shadow=False,
            is_selectable=False,
        )

        self._add_nodes(self.bases_r, self.arrows_r, show_in_hierarchy=False)

    @property
    def bounds(self):
        return self.arrows_r.bounds

    @property
    def current_bounds(self):
        return self.arrows_r.current_bounds

    @property
    def current_origins(self):
        idx = self.current_frame_id if self.origins.shape[0] > 1 else 0
        return self.origins[idx]

    @current_origins.setter
    def current_origins(self, origins):
        idx = self.current_frame_id if self.origins.shape[0] > 1 else 0
        self.origins[idx] = origins

    @property
    def current_tips(self):
        idx = self.current_frame_id if self.tips.shape[0] > 1 else 0
        return self.tips[idx]

    @current_tips.setter
    def current_tips(self, tips):
        idx = self.current_frame_id if self.tips.shape[0] > 1 else 0
        self.tips[idx] = tips

    @property
    def origins(self):
        return self._origins

    @origins.setter
    def origins(self, origins):
        self._origins = origins if len(origins.shape) == 3 else origins[np.newaxis]
        self.n_frames = self._origins.shape[0]

    @property
    def tips(self):
        return self._tips

    @tips.setter
    def tips(self, tips):
        self._tips = tips if len(tips.shape) == 3 else tips[np.newaxis]
        self.n_frames = self._tips.shape[0]

    @property
    def mid_points(self):
        return self.origins + (self.tips - self.origins) * (1 - self.p)

    def get_line_coords(self, starts, ends):
        c = np.zeros((len(self), (starts.shape[1] + ends.shape[1]), 3), dtype=starts.dtype)
        c[:, 0::2] = starts
        c[:, 1::2] = ends
        return c

    def redraw(self, **kwargs):
        self.bases_r.lines = self.get_line_coords(self.origins, self.mid_points)
        self.arrows_r.lines = self.get_line_coords(self.mid_points, self.tips)

    @Node.color.setter
    def color(self, color):
        self.material.color = color
        self.bases_r.color = color
        self.arrows_r.color = color

    def update_frames(self, origins, tips, frames):
        self.origins[frames] = origins
        self.tips[frames] = tips
        self.n_frames = self.origins.shape[0]
        self.redraw()

    def add_frames(self, origins, tips):
        if len(origins.shape) == 2:
            origins = origins[np.newaxis]
        self.origins = np.append(self.origins, origins, axis=0)

        if len(tips.shape) == 2:
            tips = tips[np.newaxis]
        self.tips = np.append(self.tips, tips, axis=0)

        self.n_frames = self.origins.shape[0]
        self.redraw()

    def remove_frames(self, frames):
        self.origins = np.delete(self.origins, frames, axis=0)
        self.tips = np.delete(self.tips, frames, axis=0)

        self.n_frames = self.origins.shape[0]
        self.redraw()



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev, Dario Mylonopoulos

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import moderngl
import numpy as np
from moderngl_window.opengl.vao import VAO

from aitviewer.scene.material import Material
from aitviewer.scene.node import Node
from aitviewer.shaders import (
    get_depth_only_program,
    get_fragmap_program,
    get_outline_program,
    get_sphere_instanced_program,
)
from aitviewer.utils.decorators import hooked
from aitviewer.utils.utils import set_lights_in_program, set_material_properties


def _create_sphere(radius=1.0, rings=16, sectors=32):
    """
    Create a sphere centered at the origin. This is a port of moderngl-window's geometry.sphere() function, but it
    returns the vertices and faces explicitly instead of directly storing them in a VAO.
    :param radius: Radius of the sphere.
    :param rings: Longitudinal resolution.
    :param sectors: Latitudinal resolution.
    :return: vertices and faces of the sphere.
    """
    R = 1.0 / (rings - 1)
    S = 1.0 / (sectors - 1)

    vertices = np.zeros((rings * sectors, 3))
    v, n = 0, 0
    for r in range(rings):
        for s in range(sectors):
            y = np.sin(-np.pi / 2 + np.pi * r * R)
            x = np.cos(2 * np.pi * s * S) * np.sin(np.pi * r * R)
            z = np.sin(2 * np.pi * s * S) * np.sin(np.pi * r * R)

            vertices[v] = np.array([x, y, z]) * radius

            v += 1
            n += 1

    faces = np.zeros([rings * sectors * 2, 3], dtype=np.int32)
    i = 0
    for r in range(rings - 1):
        for s in range(sectors - 1):
            faces[i] = np.array([r * sectors + s, (r + 1) * sectors + (s + 1), r * sectors + (s + 1)])
            faces[i + 1] = np.array([r * sectors + s, (r + 1) * sectors + s, (r + 1) * sectors + (s + 1)])
            i += 2

    return vertices, faces


class Spheres(Node):
    """Render some simple spheres."""

    def __init__(
        self,
        positions,
        radius=0.01,
        color=(0.0, 0.0, 1.0, 1.0),
        rings=16,
        sectors=32,
        icon="\u008d",
        cast_shadow=False,
        **kwargs,
    ):
        """
        Initializer.
        :param positions: A numpy array of shape (F, N, 3) or (N, 3) containing N sphere positions for F time steps.
        :param radius: Radius of the spheres.
        :param color: Color of the spheres.
        :param rings: Longitudinal resolution.
        :param sectors: Latitudinal resolution.
        """
        if len(positions.shape) == 2:
            positions = positions[np.newaxis]
        assert len(positions.shape) == 3

        # Define a default material in case there is None.
        if isinstance(color, tuple) or len(color.shape) == 1:
            kwargs["material"] = kwargs.get("material", Material(color=color, ambient=0.2))
            self.sphere_colors = kwargs["material"].color
        else:
            assert color.shape[1] == 4 and positions.shape[1] == color.shape[0]
            self.sphere_colors = color

        if "n_frames" not in kwargs:
            kwargs["n_frames"] = positions.shape[0]
        super().__init__(icon=icon, **kwargs)

        self._sphere_positions = positions
        self.radius = radius

        self.vertices, self.faces = _create_sphere(radius=1.0, rings=rings, sectors=sectors)
        self.n_vertices = self.vertices.shape[0]
        self.n_spheres = self.sphere_positions.shape[1]

        self.draw_edges = False
        self._need_upload = True

        # Render passes.
        self.outline = True
        self.fragmap = True
        self.depth_prepass = True
        self.cast_shadow = cast_shadow

    @property
    def bounds(self):
        bounds = self.get_bounds(self.sphere_positions)
        bounds[:, 0] -= self.radius
        bounds[:, 1] += self.radius
        return bounds

    @property
    def current_bounds(self):
        bounds = self.get_bounds(self.current_sphere_positions)
        bounds[:, 0] -= self.radius
        bounds[:, 1] += self.radius
        return bounds

    @property
    def vertex_colors(self):
        if len(self._sphere_colors.shape) == 1:
            return np.full((self.n_spheres * self.n_vertices, 4), self._sphere_colors)
        else:
            return np.tile(self._sphere_colors, (self.n_vertices, 1))

    def color_one(self, index, color):
        new_colors = np.tile(np.array(self.material.color), (self.n_spheres, 1))
        new_colors[index] = color
        self.sphere_colors = new_colors

    @Node.color.setter
    def color(self, color):
        self.material.color = color
        self.sphere_colors = color
        self.redraw()

    @property
    def sphere_colors(self):
        if len(self._sphere_colors.shape) == 1:
            t = np.tile(np.array(self._sphere_colors), (self.n_spheres, 1))
            return t
        else:
            return self._sphere_colors

    @sphere_colors.setter
    def sphere_colors(self, color):
        if isinstance(color, tuple):
            color = np.array(color)
        self._sphere_colors = color
        self.redraw()

    @property
    def current_sphere_positions(self):
        idx = self.current_frame_id if self.sphere_positions.shape[0] > 1 else 0
        return self.sphere_positions[idx]

    @current_sphere_positions.setter
    def current_sphere_positions(self, positions):
        assert len(positions.shape) == 2
        idx = self.current_frame_id if self.sphere_positions.shape[0] > 1 else 0
        self.sphere_positions[idx] = positions
        self.redraw()

    @property
    def sphere_positions(self):
        return self._sphere_positions

    @sphere_positions.setter
    def sphere_positions(self, pos):
        if len(pos.shape) == 2:
            pos = pos[np.newaxis]
        self._sphere_positions = pos
        self.n_frames = len(self._sphere_positions)
        self.redraw()

    def on_frame_update(self):
        self.redraw()

    def redraw(self, **kwargs):
        self._need_upload = True

    @Node.once
    def make_renderable(self, ctx: moderngl.Context):
        self.prog = get_sphere_instanced_program()

        vs_path = "sphere_instanced_positions.vs.glsl"
        self.outline_program = get_outline_program(vs_path)
        self.depth_only_program = get_depth_only_program(vs_path)
        self.fragmap_program = get_fragmap_program(vs_path)

        self.vbo_vertices = ctx.buffer(self.vertices.astype("f4").tobytes())
        self.vbo_indices = ctx.buffer(self.faces.astype("i4").tobytes())

        self.vbo_instance_position = ctx.buffer(reserve=self.n_spheres * 12)
        self.vbo_instance_color = ctx.buffer(reserve=self.n_spheres * 16)

        self.vao = VAO()
        self.vao.buffer(self.vbo_vertices, "3f4", "in_position")
        self.vao.buffer(self.vbo_instance_position, "3f4/i", "instance_position")
        self.vao.buffer(self.vbo_instance_color, "4f4/i", "instance_color")
        self.vao.index_buffer(self.vbo_indices)

    def _upload_buffers(self):
        if not self.is_renderable or not self._need_upload:
            return
        self._need_upload = False
        self.vbo_instance_position.write(self.current_sphere_positions.astype("f4").tobytes())
        if len(self._sphere_colors.shape) > 1:
            self.vbo_instance_color.write(self._sphere_colors.astype("f4").tobytes())

    def render(self, camera, **kwargs):
        self._upload_buffers()

        prog = self.prog
        prog["radius"] = self.radius
        if len(self._sphere_colors.shape) == 1:
            prog["use_uniform_color"] = True
            prog["uniform_color"] = tuple(self._sphere_colors)
        else:
            prog["use_uniform_color"] = False
        prog["draw_edges"].value = 1.0 if self.draw_edges else 0.0
        prog["win_size"].value = kwargs["window_size"]
        prog["clip_control"].value = (0, 0, 0)

        self.set_camera_matrices(prog, camera, **kwargs)
        set_lights_in_program(
            prog,
            kwargs["lights"],
            kwargs["shadows_enabled"],
            kwargs["ambient_strength"],
        )
        set_material_properties(prog, self.material)
        self.receive_shadow(prog, **kwargs)
        self.vao.render(prog, moderngl.TRIANGLES, instances=self.n_spheres)

    def render_positions(self, prog):
        if self.is_renderable:
            self._upload_buffers()
            prog["radius"] = self.radius
            self.vao.render(prog, moderngl.TRIANGLES, instances=self.n_spheres)

    def gui(self, imgui):
        _, self.radius = imgui.drag_float("Radius", self.radius, 0.01, min_value=0.001, max_value=10.0, format="%.3f")
        super().gui(imgui)

    @hooked
    def release(self):
        if self.is_renderable:
            self.vao.release()

    def update_frames(self, positions, frames):
        self.sphere_positions[frames] = positions
        self.redraw()

    def add_frames(self, positions):
        if len(positions.shape) == 2:
            positions = positions[np.newaxis]
        self.sphere_positions = np.append(self.sphere_positions, positions, axis=0)

    def remove_frames(self, frames):
        self.sphere_positions = np.delete(self.sphere_positions, frames, axis=0)
        self.redraw()


class SpheresTrail(Spheres):
    """A sequence of spheres that leaves a trail, i.e. the past spheres keep being rendered."""

    def __init__(self, positions, **kwargs):
        super().__init__(positions, **kwargs)
        self.n_frames = positions.shape[0]

    def on_frame_update(self):
        self.n_spheres = self.current_frame_id + 1
        super().on_frame_update()

    def make_renderable(self, ctx):
        super().make_renderable(ctx)
        self.on_frame_update()



from setuptools import find_packages, setup

from aitviewer import __version__

setup(
    name="aitviewer",
    description="Viewing and rendering of sequences of 3D data.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/eth-ait/aitviewer",
    version=__version__,
    author="Manuel Kaufmann, Velko Vechev, Dario Mylonopoulos",
    packages=find_packages(),
    include_package_data=True,
    keywords=[
        "viewer",
        "moderngl",
        "machine learning",
        "sequences",
        "smpl",
        "computer graphics",
        "computer vision",
        "3D",
        "meshes",
        "visualization",
    ],
    platforms=["any"],
    python_requires=">=3.7",
    install_requires=[
        "torch>=1.6.0",
        "numpy>=1.18,<2",
        "opencv-contrib-python-headless>=4.5.1.48",
        "smplx",
        "moderngl-window>=2.4.3",
        "moderngl>=5.8.2,<6",
        "PyQt5>=5.15.4",
        "imgui==2.0.0",
        "tqdm>=4.60.0",
        "trimesh>=3.9.15,<4",
        "scipy>=1.5.2,<1.8",
        "omegaconf>=2.1.1",
        "roma>=1.2.3",
        "joblib",
        "scikit-image>=0.9.0",
        "scikit-video",
        "Pillow",
        "websockets",
        "usd-core>=23.5",
    ],
)



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev, Dario Mylonopoulos

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import moderngl
import numpy as np

from aitviewer.configuration import CONFIG as C
from aitviewer.renderables.coordinate_system import CoordinateSystem
from aitviewer.renderables.lines import Lines2D
from aitviewer.renderables.plane import ChessboardPlane
from aitviewer.scene.camera import ViewerCamera
from aitviewer.scene.light import Light
from aitviewer.scene.node import Node
from aitviewer.utils.utils import (
    compute_union_of_bounds,
    compute_union_of_current_bounds,
)


class Scene(Node):
    """Generic scene node"""

    def __init__(self, **kwargs):
        """Create a scene with a name."""
        kwargs["gui_material"] = False
        super(Scene, self).__init__(**kwargs)

        # References resources in the scene
        self.lights = []
        self.camera = None

        # Scene has a reference to ctx
        self.ctx = None

        self.backface_culling = True
        self.fps = C.scene_fps
        self.background_color = C.background_color

        # Default Setup
        # If you update the number of lights, make sure to change the respective `define` statement in
        # directional_lights.glsl as well!
        # Influence of diffuse lighting is controlled globally for now, but should eventually be a material property.
        self.lights.append(
            Light.facing_origin(
                light_color=(1.0, 1.0, 1.0),
                name="Back Light",
                position=(0.0, 10.0, -15.0),
                shadow_enabled=False,
            )
        )
        self.lights.append(
            Light.facing_origin(
                light_color=(1.0, 1.0, 1.0),
                name="Front Light",
                position=(0.0, 10.0, 15.0),
            )
        )
        self.add(*self.lights)

        self.ambient_strength = 2.0

        # Scene items
        self.origin = CoordinateSystem(name="Origin", length=0.1, gui_affine=False, gui_material=False)
        self.add(self.origin)

        self.floor = ChessboardPlane(100.0, 200, (0.9, 0.9, 0.9, 1.0), (0.82, 0.82, 0.82, 1.0), name="Floor")
        self.floor.material.diffuse = 0.1
        self.add(self.floor)

        # Camera cursor rendered at the camera target.
        self.camera_target = Lines2D(
            np.array(
                [
                    [-1, 0, 0],
                    [1, 0, 0],
                    [0, -1, 0],
                    [0, 1, 0],
                    [0, 0, -1],
                    [0, 0, 1],
                ]
            ),
            color=(0.2, 0.2, 0.2, 1),
            mode="lines",
        )
        self.add(self.camera_target, show_in_hierarchy=False, enabled=False)

        # Camera trackball.
        N = 50
        t = np.empty(N * 2, dtype=np.float32)
        t[0::2] = np.linspace(0, 2.0 * np.pi, N)
        t[1::2] = np.roll(np.linspace(0, 2.0 * np.pi, N), 1)
        z = np.zeros_like(t)
        c = np.cos(t)
        s = np.sin(t)
        trackball_vertices = np.concatenate(
            (
                np.vstack((z, c, s)).T,
                np.vstack((c, z, s)).T,
                np.vstack((c, s, z)).T,
            )
        )
        trackball_colors = np.concatenate(
            (
                np.tile((1, 0, 0, 1), (N, 1)),
                np.tile((0, 0.8, 0, 1), (N, 1)),
                np.tile((0, 0, 1, 1), (N, 1)),
            )
        )

        self.trackball = Lines2D(
            trackball_vertices,
            trackball_colors,
            mode="lines",
        )
        self.add(self.trackball, show_in_hierarchy=False, enabled=False)

        self.custom_font = None
        self.properties_icon = "\u0094"

        # Currently selected object, None if no object is selected
        self.selected_object = None
        # Object shown in the property panel
        self.gui_selected_object = None

        # The scene node in the GUI is expanded at the start.
        self.expanded = True

    def render(self, **kwargs):
        # As per https://learnopengl.com/Advanced-OpenGL/Blending

        # Setup the camera target cursor for rendering.
        camera = kwargs["camera"]
        if isinstance(camera, ViewerCamera):
            # Scale target and trackball depending on distance from camera to target and fov.
            if camera.is_ortho:
                scale = camera.ortho_size * 0.8
            else:
                scale = np.radians(camera.fov) * 0.4 * np.linalg.norm(camera.target - camera.position)

            if self.camera_target.enabled:
                self.camera_target.position = camera.target
                self.camera_target.scale = scale * 0.05
            if self.trackball.enabled:
                self.trackball.position = camera.target
                self.trackball.scale = scale
        else:
            # If not a viewer camera, hide them.
            self.camera_target.scale = 0
            self.trackball.scale = 0

        # Collect all renderable nodes
        rs = self.collect_nodes()

        transparent = []

        # Draw all opaque objects first
        for r in rs:
            if not r.is_transparent():
                # Turn off backface culling if enabled for the scene
                # and requested by the current object
                if self.backface_culling and r.backface_culling:
                    self.ctx.enable(moderngl.CULL_FACE)
                else:
                    self.ctx.disable(moderngl.CULL_FACE)

                self.safe_render(r, **kwargs)
            else:
                # Otherwise append to transparent list
                transparent.append(r)

        fbo = kwargs["fbo"]

        # Draw back to front by sorting transparent objects based on distance to camera.
        # As an approximation we only sort using the origin of the object
        # which may be incorrect for large objects or objects significantly
        # offset from their origin, but works in many cases.
        for r in sorted(
            transparent,
            key=lambda x: np.linalg.norm(x.position - camera.position),
            reverse=True,
        ):
            # Turn off backface culling if enabled for the scene
            # and requested by the current object
            if self.backface_culling and r.backface_culling:
                self.ctx.enable(moderngl.CULL_FACE)
            else:
                self.ctx.disable(moderngl.CULL_FACE)

            # Render to depth buffer only
            self.ctx.depth_func = "<"

            fbo.color_mask = (False, False, False, False)
            self.safe_render_depth_prepass(r, **kwargs)
            fbo.color_mask = (True, True, True, True)

            # Render normally with less equal depth comparison function,
            # drawing only the pixels closer to the camera to avoid
            # order dependent blending artifacts
            self.ctx.depth_func = "<="
            self.safe_render(r, **kwargs)

        # Restore the default depth comparison function
        self.ctx.depth_func = "<"

    def safe_render_depth_prepass(self, r, **kwargs):
        if not r.is_renderable:
            r.make_renderable(self.ctx)
        r.render_depth_prepass(**kwargs)

    def safe_render(self, r, **kwargs):
        if not r.is_renderable:
            r.make_renderable(self.ctx)
        r.render(**kwargs)

    def make_renderable(self, ctx):
        self.ctx = ctx
        rs = self.collect_nodes(req_enabled=False)
        for r in rs:
            r.make_renderable(self.ctx)

    @property
    def bounds(self):
        return compute_union_of_bounds([n for n in self.nodes if n not in self.lights])

    @property
    def current_bounds(self):
        return compute_union_of_current_bounds(
            [n for n in self.nodes if n not in self.lights and n != self.camera_target and n != self.trackball]
        )

    @property
    def bounds_without_floor(self):
        return compute_union_of_current_bounds([n for n in self.nodes if n not in self.lights and n != self.floor])

    def auto_set_floor(self):
        """Finds the minimum lower bound in the y coordinate from all the children bounds and uses that as the floor"""
        if self.floor is not None and len(self.nodes) > 0:
            self.floor.position[1] = self.current_bounds[1, 0]
            self.floor.update_transform()

    def auto_set_camera_target(self):
        """Sets the camera target to the average of the center of all objects in the scene"""
        centers = []
        for n in self.nodes:
            if n not in self.lights:
                centers.append(n.current_center)

        if isinstance(self.camera, ViewerCamera) and len(centers) > 0:
            self.camera.target = np.array(centers).mean(0)

    @property
    def light_mode(self):
        return self._light_mode

    @light_mode.setter
    def light_mode(self, mode):
        if mode == "default":
            self._light_mode = mode
            self.ambient_strength = 2.0
            for l in self.lights:
                l.strength = 1.0
        elif mode == "dark":
            self._light_mode = mode
            self.ambient_strength = 0.4
            for l in self.lights:
                l.strength = 1.0
        elif mode == "diffuse":
            self._light_mode = mode
            self.ambient_strength = 1.0
            for l in self.lights:
                l.strength = 2.0
        else:
            raise ValueError(f"Invalid light mode: {mode}")

    def collect_nodes(self, req_enabled=True, obj_type=Node):
        nodes = []

        # Use head recursion in order to collect the most deep nodes first
        # These are likely to be nodes which should be rendered first (i.e. transparent parent nodes should be last)
        def rec_collect_nodes(nn):
            if not req_enabled or nn.enabled:
                if isinstance(nn, obj_type):
                    nodes.append(nn)
                for n_child in nn.nodes:
                    rec_collect_nodes(n_child)

        for n in self.nodes:
            rec_collect_nodes(n)
        return nodes

    def get_node_by_name(self, name):
        assert name != ""
        ns = self.collect_nodes()
        for n in ns:
            if n.name == name:
                return n
        return None

    def get_node_by_uid(self, uid):
        ns = self.collect_nodes()
        for n in ns:
            if n.uid == uid:
                return n
        return None

    def select(self, obj, selected_node=None, selected_instance=None, selected_tri_id=None):
        """Set 'obj' as the selected object"""
        self.selected_object = obj
        if isinstance(obj, Node):
            self.selected_object.on_selection(selected_node, selected_instance, selected_tri_id)
        # Always keep the last selected object in the property panel
        if obj is not None:
            self.gui_selected_object = obj

    def is_selected(self, obj):
        """Returns true if obj is currently selected"""
        return obj == self.selected_object

    def gui_selected(self, imgui):
        """GUI to edit the selected node"""
        if self.gui_selected_object:
            s = self.gui_selected_object

            # Custom GUI Elements
            imgui.indent(22)
            imgui.push_font(self.custom_font)
            imgui.text(f"{s.icon} {s.name}")
            imgui.pop_font()

            # Modes
            if hasattr(s, "gui_modes") and len(s.gui_modes) > 1:
                imgui.push_font(self.custom_font)
                imgui.spacing()
                for i, (gm_key, gm_val) in enumerate(s.gui_modes.items()):
                    if s.selected_mode == gm_key:
                        imgui.push_style_color(imgui.COLOR_BUTTON, 0.26, 0.59, 0.98, 1.0)
                    mode_clicked = imgui.button(f" {gm_val['icon']}{gm_val['title']} ")
                    if s.selected_mode == gm_key:
                        imgui.pop_style_color()
                    if mode_clicked:
                        s.selected_mode = gm_key
                    if i != len(s.gui_modes) - 1:
                        imgui.same_line()
                imgui.pop_font()

                # Mode specific GUI
                imgui.spacing()
                if "fn" in s.gui_modes[s.selected_mode]:
                    s.gui_modes[s.selected_mode]["fn"](imgui)

            # Custom GUI (i.e. Camera specific params)
            s.gui(imgui)
            imgui.unindent()

            # General GUI elements
            if hasattr(s, "gui_controls"):
                imgui.spacing()
                imgui.spacing()
                imgui.spacing()
                for i, (gc_key, gc_val) in enumerate(s.gui_controls.items()):
                    if not gc_val["is_visible"]:
                        continue
                    imgui.begin_group()
                    imgui.push_font(self.custom_font)
                    imgui.text(f"{gc_val['icon']}")
                    imgui.pop_font()
                    imgui.end_group()
                    imgui.same_line(spacing=8)
                    imgui.begin_group()
                    gc_val["fn"](imgui)
                    imgui.end_group()
                    imgui.spacing()
                    imgui.spacing()
                    imgui.spacing()

    def gui(self, imgui):
        imgui.text(f"FPS: {self.fps:.1f}")
        # Background color
        uc, color = imgui.color_edit4("Background", *self.background_color)
        if uc:
            self.background_color = color

        _, self.ambient_strength = imgui.drag_float(
            "Ambient strength",
            self.ambient_strength,
            0.01,
            min_value=0.0,
            max_value=10.0,
            format="%.2f",
        )

    def gui_editor(self, imgui, viewports, viewport_mode):
        """GUI to control scene settings."""
        # Also include the camera GUI in the scene node.
        self.gui_camera(imgui, viewports, viewport_mode)
        imgui.spacing()
        imgui.separator()
        imgui.spacing()
        self.gui_hierarchy(imgui, [self])
        imgui.spacing()
        imgui.separator()
        imgui.spacing()
        imgui.spacing()
        self.gui_selected(imgui)

    def gui_camera(self, imgui, viewports, viewport_mode):
        # Label for the viewport position, if using multiple viewports.
        if viewport_mode == "single":
            suffixes = [""]
        elif viewport_mode == "split_v":
            suffixes = ["(left)", "(right)"]
        elif viewport_mode == "split_h":
            suffixes = ["(top)", "(bottom)"]
        else:
            suffixes = [
                "(top left)",
                "(top right)",
                "(bottom left)",
                "(bottom right)",
            ]

        for suffix, v in zip(suffixes, viewports):
            camera = v.camera

            # Camera GUI
            imgui.push_font(self.custom_font)
            imgui.push_style_var(imgui.STYLE_FRAME_PADDING, (0, 2))

            flags = imgui.TREE_NODE_LEAF | imgui.TREE_NODE_FRAME_PADDING
            if self.is_selected(camera):
                flags |= imgui.TREE_NODE_SELECTED

            if isinstance(camera, ViewerCamera):
                name = camera.name
            else:
                name = f"Camera: {camera.name}"

            camera_expanded = imgui.tree_node(f"{camera.icon}  {name}##tree_node_r_camera", flags)
            if imgui.is_item_clicked():
                self.select(camera)

            # Add a right justified label with the viewport position.
            if suffix:
                imgui.same_line()
                pos = imgui.get_cursor_pos_x()
                avail = imgui.get_content_region_available()[0] - 3
                # Only put the label if enough space is available for it.
                if avail > imgui.calc_text_size(suffix)[0]:
                    imgui.same_line()
                    avail -= imgui.calc_text_size(suffix)[0]
                    imgui.set_cursor_pos_x(pos + avail)
                    imgui.text(suffix)
                else:
                    # Put an empty label to go to the next line.
                    imgui.text("")

            imgui.pop_style_var()
            imgui.pop_font()
            if camera_expanded:
                imgui.tree_pop()

    def gui_hierarchy(self, imgui, rs):
        # Nodes GUI
        for r in rs:
            # Skip nodes that shouldn't appear in the hierarchy.
            if not r.show_in_hierarchy:
                continue

            # Visibility
            curr_enabled = r.enabled
            if not curr_enabled:
                imgui.push_style_color(imgui.COLOR_TEXT, 1.0, 1.0, 1.0, 0.4)

            # Title
            imgui.push_font(self.custom_font)
            imgui.push_style_var(imgui.STYLE_FRAME_PADDING, (0, 2))

            flags = imgui.TREE_NODE_OPEN_ON_ARROW | imgui.TREE_NODE_FRAME_PADDING
            if r.expanded:
                flags |= imgui.TREE_NODE_DEFAULT_OPEN
            if self.is_selected(r):
                flags |= imgui.TREE_NODE_SELECTED
            if not any(c.show_in_hierarchy for c in r.nodes):
                flags |= imgui.TREE_NODE_LEAF
            r.expanded = imgui.tree_node("{} {}##tree_node_{}".format(r.icon, r.name, r.unique_name), flags)
            if imgui.is_item_clicked():
                self.select(r)

            imgui.pop_style_var()
            imgui.pop_font()

            if r != self:
                # Aligns checkbox to the right side of the window
                # https://github.com/ocornut/imgui/issues/196
                imgui.same_line(position=imgui.get_window_content_region_max().x - 25)
                eu, enabled = imgui.checkbox("##enabled_r_{}".format(r.unique_name), r.enabled)
                if eu:
                    r.enabled = enabled

            if r.expanded:
                # Recursively render children nodes
                self.gui_hierarchy(imgui, r.nodes)
                imgui.tree_pop()

            if not curr_enabled:
                imgui.pop_style_color(1)

    def add_light(self, light):
        self.lights.append(light)

    @property
    def n_lights(self):
        return len(self.lights)

    @property
    def n_frames(self):
        n_frames = 1
        ns = self.collect_nodes(req_enabled=False)
        for n in ns:
            if n._enabled_frames is None:
                n_frames = max(n_frames, n.n_frames)
            else:
                n_frames = max(n_frames, n._enabled_frames.shape[0])
        return n_frames

    def render_outline(self, *args, **kwargs):
        # No outline when the scene node is selected
        return



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev, Dario Mylonopoulos

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np
from skimage import measure

from aitviewer.renderables.bounding_boxes import BoundingBoxes
from aitviewer.renderables.lines import Lines
from aitviewer.renderables.meshes import Meshes
from aitviewer.scene.node import Node
from aitviewer.utils.decorators import hooked


class SDF(Node):
    """
    Renderable that can be used to draw level sets of a dense SDF volume meshed using marching cubes.

    This renderable internally uses the marching cubes algorithm from skimage.
    For a faster marching cubes implementation see the Volume renderable.
    """

    def __init__(
        self,
        volume,
        size=(1, 1, 1),
        level=0.0,
        color=(0.7, 0.7, 0.7, 1.0),
        level_sets=None,
        level_set_colors=None,
        mc_step_size=1,
        **kwargs,
    ):
        """Initializer.
        :param volume: np array of shape (X, Y, Z) of signed distance values
        :param size: size of the volume in local units.
        :param level: the level set used for the main mesh.
        :param color: color of the main mesh.
        :param level_sets: a list or array of additional level set values to display.
        :param level_set_colors: a list or array of shape (L, 4) of the same length as
            the level_set parameter with colors to use for the additional level sets.
        :param mc_step_size: step size used for marching cubes.
        :param **kwargs: arguments forwarded to the Node constructor.
        """
        assert len(volume.shape) == 3 and len(size) == 3
        kwargs["gui_material"] = False
        super().__init__(**kwargs)

        self.volume = volume
        self.size = np.array((size), np.float32)

        # Mesh.
        verts, faces, normals, _ = measure.marching_cubes(
            volume, level, spacing=self.size / (np.array(self.volume.shape) - 1.0), step_size=mc_step_size
        )
        self.mesh = Meshes(verts, faces, vertex_normals=-normals, color=color, name="Mesh")

        # Level sets.
        self.level_sets: list[Meshes] = []
        if level_sets is not None:
            if level_set_colors is not None:
                assert len(level_sets) == len(level_set_colors)

            for i, s in enumerate(level_sets):
                verts, faces, normals, _ = measure.marching_cubes(
                    volume, s, spacing=self.size / (np.array(self.volume.shape) - 1.0), step_size=mc_step_size
                )
                shell = Meshes(verts, faces, vertex_normals=-normals, name=f"Level {s:.03f}", cast_shadow=False)
                if level_set_colors is not None:
                    shell.color = tuple(level_set_colors[i])

                shell.clip_control = np.array((1, 1, 1))
                shell.clip_value = self.size.copy()
                shell.backface_culling = False

                self.level_sets.append(shell)

        # Bounding box.
        self.bounding_box = BoundingBoxes.from_min_max_diagonal(
            np.array([[0.0, 0.0, 0.0]]),
            np.array([self.size], dtype=np.float32),
            color=(0, 0, 0, 1),
            name="Bounding Box",
            gui_affine=False,
        )

        # Clip plane lines.
        self.clip_lines = []
        for i, axis in enumerate(["X", "Y", "Z"]):
            s0 = self.size[(i + 0) % 3]
            s1 = self.size[(i + 1) % 3]
            s2 = self.size[(i + 2) % 3]
            lines = np.array(
                (
                    [s0, 0, 0],
                    [s0, s1, 0],
                    [s0, s1, s2],
                    [s0, 0, s2],
                    [s0, 0, 0],
                ),
                dtype=np.float32,
            )
            lines = np.roll(lines, axis=1, shift=(0, i))
            color = np.array([0, 0, 0, 1])
            color[i] = 1
            self.clip_lines.append(Lines(lines, cast_shadow=False, color=color, name=f"Clip {axis}", gui_affine=False))

        # Group clip planes under a common parent node.
        self.clip_planes_node = Node("Clip planes", gui_affine=False, gui_material=False)
        self.clip_planes_node.add(*self.clip_lines)

        # Group level sets under a common parent node.
        self.level_sets_node = Node("Level sets", gui_material=False)
        self.level_sets_node.add(*self.level_sets)

        # Add all children nodes.
        self.add(self.mesh, self.level_sets_node, self.bounding_box, self.clip_planes_node)

        # Initialze clip extents and reverse booleans.
        self._clip_extents = np.array([1.0, 1.0, 1.0])
        self._clip_reversed = np.array([False, False, False])

    @classmethod
    def with_level_sets(
        cls,
        volume,
        inside_levels=None,
        inside_color_start=(0.6, 0.6, 0.7, 1.0),
        inside_color_end=(0.2, 0.2, 0.7, 1.0),
        outside_levels=None,
        outside_color_start=(0.7, 0.6, 0.6, 1.0),
        outside_color_end=(0.7, 0.2, 0.2, 1.0),
        **kwargs,
    ) -> "SDF":
        """
        Crates an SDF object with additional level sets inside and/or outside the mesh.
        Additionally a starting and end color can be specified for inside and outside level sets,
        level sets will be colored interpolating between these two values.

        :param volume: np array of shape (X, Y, Z) of signed distance values.
        :param inside_levels: list or array of level set values for the inside of the mesh.
        :param inside_color_start: starting color for the inside level sets.
        :param inside_color_end: end color for the inside level sets.
        :param outside_levels: list or array of level set values for the outside of the mesh.
        :param outside_color_start: starting color for the outside level sets.
        :param outside_color_end: end color for the outside level sets.
        :param **kwargs: forwarded to the SDF constructor.

        :return: SDF renderable.
        """
        level_sets = []
        level_set_colors = []
        if inside_levels is not None:
            inside_levels = np.array(inside_levels)
            inside_color_start = np.array(inside_color_start, dtype=np.float32)
            inside_color_end = np.array(inside_color_end, dtype=np.float32)
            inside_colors = np.linspace(inside_color_start, inside_color_end, inside_levels.shape[0])
            level_sets.append(inside_levels)
            level_set_colors.append(inside_colors)

        if outside_levels is not None:
            outside_levels = np.array(outside_levels)
            outside_color_start = np.array(outside_color_start, dtype=np.float32)
            outside_color_end = np.array(outside_color_end, dtype=np.float32)
            outside_colors = np.linspace(outside_color_start, outside_color_end, outside_levels.shape[0])
            level_sets.append(outside_levels)
            level_set_colors.append(outside_colors)

        level_sets = np.hstack(level_sets)
        level_set_colors = np.vstack(level_set_colors)
        return cls(volume, level_sets=level_sets, level_set_colors=level_set_colors, **kwargs)

    @property
    def bounds(self):
        return self.current_bounds

    @property
    def current_bounds(self):
        min = np.array([0, 0, 0])
        max = self.size
        return self.get_bounds(np.vstack((min, max)))

    @property
    def clip_extents(self):
        """Property controlling the clip extents used to clip the additional level sets."""
        return self._clip_extents

    @clip_extents.setter
    def clip_extents(self, v):
        """Setter of property controlling the clip extents used to clip the additional level sets."""
        assert len(v) == 3
        for i in range(3):
            if self.clip_extents[i] == v[i]:
                continue
            self._clip_extents[i] = v[i]
            self.clip_lines[i].lines[:, :, i] = v[i] * self.size[i]
            self.clip_lines[i].redraw()
            for s in self.level_sets:
                s.clip_value[i] = v[i] * self.size[i]

    @property
    def clip_reversed(self):
        """Property controlling if the clip extents used to clip the additional level sets are reversed."""
        return self._clip_reversed

    @clip_reversed.setter
    def clip_reversed(self, v):
        """Setter of property controlling if the clip extents used to clip the additional level sets are reversed."""
        assert len(v) == 3
        self._clip_reversed = np.array(v)
        for i in range(3):
            for s in self.shells:
                s.clip_control[i] = -1 if self._clip_reversed[i] else 1

    @hooked
    def gui(self, imgui):
        u, val = imgui.drag_float3(
            f"Clip extents##{self.uid}", *self.clip_extents, 2e-3, format="%.3f", min_value=0.0, max_value=1.0
        )
        if u:
            self.clip_extents = val
        imgui.spacing()
        imgui.text("Reverse axis: ")
        imgui.same_line()
        for i, axis in enumerate(["X", "Y", "Z"]):
            _, self.clip_reversed[i] = imgui.checkbox(f"{axis} ", self.clip_reversed[i])
            if i != 2:
                imgui.same_line()
            for s in self.level_sets:
                s.clip_control[i] = -1 if self._clip_reversed[i] else 1

    def render_outline(self, *args, **kwargs):
        # Disable outline when the node itself is selected.
        pass



"""
Copyright (C) 2022  ETH Zurich, Manuel Kaufmann, Velko Vechev, Dario Mylonopoulos

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np

from aitviewer.renderables.rigid_bodies import RigidBodies


class CoordinateSystem(RigidBodies):
    """
    Render a coordinate system using shaded cylinders.
    """

    def __init__(self, length=1.0, icon="\u008a", **kwargs):
        r = length / 50
        l = length
        super(CoordinateSystem, self).__init__(
            np.array([[[0.0, 0.0, 0.0]]]),
            np.eye(3)[np.newaxis, np.newaxis],
            radius=r,
            length=l,
            icon=icon,
            **kwargs,
        )



"""
Biovision Hierarchy (BVH) reader for python with numpy pose output.

Modified from:
https://github.com/dabeschte/npybvh/tree/master


MIT License

Copyright (c) 2019 Mathias Parger

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import re

import numpy as np


class BvhJoint:
    def __init__(self, name, parent):
        self.name = name
        self.parent = parent
        self.offset = np.zeros(3)
        self.channels = []
        self.children = []

    def add_child(self, child):
        self.children.append(child)

    def __repr__(self):
        return self.name

    def position_animated(self):
        return any([x.endswith("position") for x in self.channels])

    def rotation_animated(self):
        return any([x.endswith("rotation") for x in self.channels])


class Bvh:
    def __init__(self):
        self.joints = {}
        self.root = None
        self.keyframes = None
        self.frames = 0
        self.fps = 0

    def _parse_hierarchy(self, text):
        lines = re.split("\\s*\\n+\\s*", text)

        joint_stack = []

        for line in lines:
            words = re.split("\\s+", line)
            instruction = words[0]

            if instruction == "JOINT" or instruction == "ROOT":
                parent = joint_stack[-1] if instruction == "JOINT" else None
                joint = BvhJoint(words[1], parent)
                self.joints[joint.name] = joint
                if parent:
                    parent.add_child(joint)
                joint_stack.append(joint)
                if instruction == "ROOT":
                    self.root = joint
            elif instruction == "CHANNELS":
                for i in range(2, len(words)):
                    joint_stack[-1].channels.append(words[i])
            elif instruction == "OFFSET":
                for i in range(1, len(words)):
                    joint_stack[-1].offset[i - 1] = float(words[i])
            elif instruction == "End":
                parent = joint_stack[-1]
                joint = BvhJoint(f"{parent.name}_end{len(parent.children)}", parent)
                parent.add_child(joint)
                joint_stack.append(joint)
                self.joints[joint.name] = joint
            elif instruction == "}":
                joint_stack.pop()

    def _add_pose_recursive(self, joint, offset, poses):
        pose = joint.offset + offset
        poses.append(pose)

        for c in joint.children:
            self._add_pose_recursive(c, pose, poses)

    def plot_hierarchy(self):
        import matplotlib.pyplot as plt

        poses = []
        self._add_pose_recursive(self.root, np.zeros(3), poses)
        pos = np.array(poses)

        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")
        ax.scatter(pos[:, 0], pos[:, 2], pos[:, 1])
        ax.set_xlim(-30, 30)
        ax.set_ylim(-30, 30)
        ax.set_zlim(-30, 30)
        plt.show()

    def parse_motion(self, text):
        lines = re.split("\\s*\\n+\\s*", text)

        frame = 0
        for line in lines:
            if line == "":
                continue
            words = re.split("\\s+", line)

            if line.startswith("Frame Time:"):
                self.fps = round(1 / float(words[2]))
                continue
            if line.startswith("Frames:"):
                self.frames = int(words[1])
                continue

            if self.keyframes is None:
                self.keyframes = np.empty((self.frames, len(words)), dtype=np.float32)

            for angle_index in range(len(words)):
                self.keyframes[frame, angle_index] = float(words[angle_index])

            frame += 1

    def parse_string(self, text):
        hierarchy, motion = text.split("MOTION")
        self._parse_hierarchy(hierarchy)
        self.parse_motion(motion)

    def _extract_rotation(self, frame_pose, index_offset, joint):
        M_rotation = np.eye(3)
        for channel in joint.channels:
            if channel.endswith("position"):
                continue
            local_rotation = frame_pose[index_offset]
            index_offset += 1
            a = np.deg2rad(local_rotation)
            s = np.sin(a)
            c = np.cos(a)
            if channel == "Xrotation":
                M_channel = np.array([[1, 0, 0], [0, c, -s], [0, s, c]])
            elif channel == "Yrotation":
                M_channel = np.array([[c, 0, s], [0, 1, 0], [-s, 0, c]])
            elif channel == "Zrotation":
                M_channel = np.array([[c, -s, 0], [s, c, 0], [0, 0, 1]])
            else:
                raise Exception(f"Unknown channel {channel}")
            M_rotation = M_rotation.dot(M_channel)

        return M_rotation, index_offset

    def _extract_position(self, joint, frame_pose, index_offset):
        offset_position = np.zeros(3)
        for channel in joint.channels:
            if channel.endswith("rotation"):
                continue
            if channel == "Xposition":
                offset_position[0] = frame_pose[index_offset]
            elif channel == "Yposition":
                offset_position[1] = frame_pose[index_offset]
            elif channel == "Zposition":
                offset_position[2] = frame_pose[index_offset]
            else:
                raise Exception(f"Unknown channel {channel}")
            index_offset += 1

        return offset_position, index_offset

    def _recursive_apply_frame(self, joint, frame_pose, index_offset, p, r, M_parent, p_parent):
        if joint.position_animated():
            offset_position, index_offset = self._extract_position(joint, frame_pose, index_offset)
        else:
            offset_position = np.zeros(3)

        if len(joint.channels) == 0:
            joint_index = list(self.joints.values()).index(joint)
            p[joint_index] = p_parent + M_parent.dot(joint.offset)
            r[joint_index] = M_parent
            return index_offset

        if joint.rotation_animated():
            M_rotation, index_offset = self._extract_rotation(frame_pose, index_offset, joint)
        else:
            M_rotation = np.eye(3)

        M = M_parent.dot(M_rotation)
        position = p_parent + M_parent.dot(joint.offset) + offset_position

        rotation = M
        joint_index = list(self.joints.values()).index(joint)
        p[joint_index] = position
        r[joint_index] = rotation

        for c in joint.children:
            index_offset = self._recursive_apply_frame(c, frame_pose, index_offset, p, r, M, position)

        return index_offset

    def frame_pose(self, frame):
        p = np.empty((len(self.joints), 3))
        r = np.empty((len(self.joints), 3, 3))
        frame_pose = self.keyframes[frame]
        M_parent = np.zeros((3, 3))
        M_parent[0, 0] = 1
        M_parent[1, 1] = 1
        M_parent[2, 2] = 1
        self._recursive_apply_frame(self.root, frame_pose, 0, p, r, M_parent, np.zeros(3))

        return p, r

    def all_frame_poses(self):
        p = np.empty((self.frames, len(self.joints), 3))
        r = np.empty((self.frames, len(self.joints), 3, 3))

        for frame in range(len(self.keyframes)):
            p[frame], r[frame] = self.frame_pose(frame)

        return p, r
